from __future__ import annotations

import logging
import os
import queue
import random
import threading
import time
import typing as t
from concurrent.futures import Future

import pika
import pika.channel
import pika.exceptions
from pika.spec import Basic, BasicProperties

from .base import RabbitPublisherStatus

if t.TYPE_CHECKING:
    from pika.channel import Channel
    from pika.frame import Method
    from pika.spec import Exchange
    from pika.spec import Queue as pikaQueue

log = logging.getLogger(__name__)


class ResultPublisher(threading.Thread):
    """
    Publish results to the AMQP service, per the routing information in `queue_info`.
    """

    def __init__(
        self,
        *,
        queue_info: dict,
        poll_period_s: float = 0.5,
        connect_attempt_limit: int = 5,
        channel_close_window_s: int = 10,
        channel_close_window_limit: int = 3,
    ):
        """
        :param queue_info: Pika connection parameters to connect to RabbitMQ;
            typically as returned from the web-service by the web-service
        :param poll_period_s: [default: 0.5] how frequently to check for and
            handle events.  For example, if the thread should stop or if there
            are results to send along.
        :param connect_attempt_limit: Number of connection attempts to fail before
            giving up.  The connection counter will reset to 0 after the connection
            is sustained for 60s, so transient network errors should not build up
            to a future failure.)
        :param channel_close_window_s: Window of time to count channel close events
        :param channel_close_window_limit: Limit of channel close events (within
            ``channel_close_window_s``) before shutting down the thread.
        """
        self.queue_info = queue_info

        # how often to check for work; every `poll_period_s`, the `_event_watcher`
        # method will handle any outstanding work.
        self.poll_period_s = max(0.01, poll_period_s)

        # how many times to attempt connection before giving up and shutting
        # down the thread
        self.connect_attempt_limit = max(1, connect_attempt_limit)
        self._connection_tries = 0  # count of connection events; reset on success

        # invalid until set in _on_queue_verified
        self._connected_at: float | None = None

        # list of times that channel was last closed; see _on_channel_closed
        self._channel_closes: list[float] = []

        # how long a time frame to keep previous channel close times
        self.channel_close_window_s = channel_close_window_s

        # how many times allowed to retry opening a channel in the above time
        # window before giving up and shutting down the thread
        self.channel_close_window_limit = channel_close_window_limit

        self._mq_conn: pika.SelectConnection | None = None
        self._thread_id = 0  # 0 == not yet started
        self._mq_chan: Channel | None = None
        self._awaiting_confirmation: dict[int, Future] = {}
        self._outstanding: queue.Queue[tuple[Future, bytes]] = queue.Queue()
        self._total_published = 0
        self._delivery_tag_index = 0
        self._stop_event = threading.Event()

        # start closed ("connected" after connect)
        self.status = RabbitPublisherStatus.closed

        publish_kw = dict(**self.queue_info["queue_publish_kwargs"])
        if "properties" in publish_kw:
            publish_kw["properties"] = BasicProperties(**publish_kw["properties"])
        self._publish_kwargs = publish_kw

        super().__init__()

    def __repr__(self):
        return "{}<{}; o:{:,}; t:{:,}>".format(
            type(self).__name__,
            "✓" if self.status == RabbitPublisherStatus.connected else "✗",
            self._outstanding.qsize(),
            self._total_published,
        )

    def run(self) -> None:
        log.debug("%r thread begins", self)
        self._thread_id = threading.get_ident()

        while (
            not self._stop_event.is_set()
            and self._connection_tries < self.connect_attempt_limit
        ):
            if self._mq_conn or self._connection_tries:
                wait_for = random.uniform(0.5, 10)
                msg = f"%r reconnecting in {wait_for:.1f}s."
                log.debug(msg, self)
                if self._connection_tries == self.connect_attempt_limit - 1:
                    log.warning(f"{msg}  (final attempt)", self)

                if self._stop_event.wait(wait_for):
                    break

            self._connection_tries += 1
            try:
                log.debug(
                    "%r Opening connection to AMQP service.  Attempt: %s (of %s)",
                    self,
                    self._connection_tries,
                    self.connect_attempt_limit,
                )
                self._connected_at = None
                self._mq_conn = self._connect()
                self._event_watcher()
                self._mq_conn.ioloop.start()  # Reminder: blocks

            except Exception:
                log.exception("%r Unhandled error; event loop stopped", self)
        self._stop_event.set()

        log.debug("%r thread ends.", self)

    def stop(self, block=True, timeout=5) -> None:
        if threading.get_ident() == self._thread_id:
            log.error("Programming error: invalid attempt to self-shutdown thread.")
            return

        self._stop_event.set()
        if block:
            self.join(timeout=timeout)

    def _connect(self) -> pika.SelectConnection:
        pika_params = pika.URLParameters(self.queue_info["connection_url"])
        return pika.SelectConnection(
            pika_params,
            on_open_callback=self._on_connection_open,
            on_close_callback=self._on_connection_closed,
            on_open_error_callback=self._on_open_failed,
        )

    def _on_open_failed(self, _mq_conn: pika.BaseConnection, exc: str | Exception):
        assert self._mq_conn is not None, "Strictly called _by_ ioloop"
        count = f"[attempt {self._connection_tries} (of {self.connect_attempt_limit})]"
        pid = f"(pid: {os.getpid()})"
        exc_text = f"Failed to open connection - ({exc.__class__.__name__}) {exc}"
        msg = f"{count} {pid} {exc_text}"
        log.debug(msg)

        if not (self._connection_tries < self.connect_attempt_limit):
            log.warning(msg)
            if not isinstance(exc, Exception):
                exc = Exception(str(exc))
            self._cancellation_reason = exc
        self._mq_conn.ioloop.stop()

    def _on_connection_closed(self, _mq_conn: pika.BaseConnection, exc: Exception):
        assert self._mq_conn is not None, "Strictly called _by_ ioloop"
        log.debug("%r Connection closed: %s", self, exc)
        self.status = RabbitPublisherStatus.closed
        self._mq_conn.ioloop.stop()

    def _on_connection_open(self, _mq_conn: pika.BaseConnection):
        log.debug("%r Connection established; creating channel", self)
        self._open_channel()

    def _open_channel(self):
        if self._mq_conn.is_open:
            self._mq_conn.channel(on_open_callback=self._on_channel_open)

    def _on_channel_open(self, mq_chan: Channel):
        self._mq_chan = mq_chan
        self._delivery_tag_index = 0

        mq_chan.add_on_close_callback(self._on_channel_closed)
        mq_chan.add_on_cancel_callback(self._on_consumer_cancelled)

        log.debug(
            "Channel %s opened (%s); testing.",
            mq_chan.channel_number,
            mq_chan.connection.params,
        )
        mq_chan.exchange_declare(
            passive=True,
            exchange=self.queue_info["exchange"],
            callback=self._on_exchange_verified,
        )

    def _on_consumer_cancelled(self, frame: Method[Basic.CancelOk]):
        log.info("Consumer cancelled remotely, shutting down: %r", frame)
        if self._mq_chan:
            self._mq_chan.close()

    def _on_channel_closed(self, mq_chan: Channel, exc: Exception):
        self.status = RabbitPublisherStatus.closed
        self._connected_at = None
        now = time.monotonic()
        then = now - self.channel_close_window_s
        self._channel_closes = [cc for cc in self._channel_closes if cc > then]
        self._channel_closes.append(now)
        if len(self._channel_closes) < self.channel_close_window_limit:
            if self._stop_event.is_set():
                return
            msg = f"{self} Channel closed  [{mq_chan}\n  ({exc})]"
            log.debug(msg, exc_info=exc)
            log.warning(msg)
            mq_chan.connection.ioloop.call_later(1, self._open_channel)

        else:
            log.error(
                f"{self} Unable to sustain channel after {len(self._channel_closes)}"
                f" attempts in {self.channel_close_window_limit} seconds. ({exc})"
            )
            self._stop_event.set()

    def _on_exchange_verified(self, _frame: Method[Exchange.DeclareOk]):
        assert self._mq_chan is not None
        self._mq_chan.queue_declare(
            passive=True,
            queue=self.queue_info["queue"],
            callback=self._on_queue_verified,
        )

    def _on_queue_verified(self, _frame: Method[pikaQueue.DeclareOk]):
        assert self._mq_chan is not None
        self._mq_chan.confirm_delivery(self._on_delivery)
        self.status = RabbitPublisherStatus.connected
        self._connected_at = time.time()

    def _on_delivery(self, frame: Method):
        """
        Complete futures associated frame delivery tag.

        In a SelectConnection setup, the AMQP server may send more than one
        confirmation in a single frame.  In that case, `frame.method.multiple`
        will be `True`, and means that all outstanding message with a lower
        delivery tag index value is also Ack or Nacked.
        """
        is_ack = frame.method.INDEX == Basic.Ack.INDEX
        is_nack = frame.method.INDEX == Basic.Nack.INDEX
        is_multiple = frame.method.multiple
        frame_dtag = frame.method.delivery_tag

        f = self._awaiting_confirmation.pop(frame_dtag)
        if is_ack:
            f.set_result(None)

            if is_multiple:
                for dtag in list(self._awaiting_confirmation.keys()):
                    if dtag < frame_dtag:
                        f = self._awaiting_confirmation.pop(dtag)
                        f.set_result(None)
        elif is_nack:
            exc = Exception(f"Failed to publish delivery_tag {frame_dtag}")
            f.set_exception(exc)
            if is_multiple:
                for dtag in list(self._awaiting_confirmation.keys()):
                    if dtag < frame_dtag:
                        f = self._awaiting_confirmation.pop(dtag)
                        exc = Exception(f"Failed to publish delivery_tag {dtag}")
                        f.set_exception(exc)

    def _event_watcher(self):
        if self._stop_event.is_set():
            log.info("%r Shutting down due to stop event set.")
            self._stop_ioloop()
            return

        if self._connection_tries and self._connected_at:
            # we're connected ...
            if time.time() - self._connected_at > 60:
                # ... and connection stable for 60s; good to reset connection tries
                self._connection_tries = 0
                log.debug("%r Connection deemed stable; resetting connection tally")

        try:
            if self.status != RabbitPublisherStatus.connected:
                log.debug("%r not yet connected; holding enqueued messages")
                raise queue.Empty()

            while True:
                f, msg_bytes = self._outstanding.get(block=False)
                try:
                    self._mq_chan.basic_publish(body=msg_bytes, **self._publish_kwargs)
                    self._delivery_tag_index += 1
                    self._awaiting_confirmation[self._delivery_tag_index] = f
                    self._total_published += 1
                except pika.exceptions.AMQPError as err:
                    exc = RuntimeError(
                        "Failed to publish result (delivery_tag:"
                        f" {self._delivery_tag_index}"
                    )
                    exc.__cause__ = err
                    f.set_exception(exc)
                    log.error(
                        f"Unable to deliver result to exchange.\n"
                        f"  Error text: {err}\n"
                        f"  kwargs: {self._publish_kwargs}\n"
                        f"  message: {msg_bytes!r}"
                    )
                    continue

        except queue.Empty:
            pass

        finally:
            self._mq_conn.ioloop.call_later(self.poll_period_s, self._event_watcher)

    def _stop_ioloop(self):
        """
        Gracefully stop the ioloop.

        In an effort play nice with upstream, attempt to follow the AMQP protocol
        by closing the channel and connections gracefully.  This method will
        rearm itself while the connection is still open, continually working
        toward eventually and gracefully stopping the connection, before finally
        stopping the ioloop.
        """
        if self._mq_conn:
            self._mq_conn.ioloop.call_later(0.1, self._stop_ioloop)
            if self._mq_conn.is_open:
                if self._mq_chan:
                    if self._mq_chan.is_open:
                        self._mq_chan.close()
                    elif self._mq_chan.is_closed:
                        self._mq_chan = None
                else:
                    self._mq_conn.close()
            elif self._mq_conn.is_closed:
                self._mq_conn.ioloop.stop()
                self._mq_conn = None

    def publish(self, message: bytes) -> Future[None]:
        """
        Enqueue a message to be published over the AMQP channel.  Returns a future
        that can be consulted to determine when the message is acknowledged by the
        AMQP broker.

        :param message: The raw message to send to the broker.

        :raises: ValueError is the publishing thread is not running
        """
        if not self.is_alive():
            raise ValueError("cannot publish(); thread not started")

        f: Future[None] = Future()
        self._outstanding.put((f, message))
        return f



import itertools
import sys
import time
import types
import typing as t


def create_traceback(start: int = 0) -> types.TracebackType:
    """
    Dynamically create a traceback.

    Builds a traceback from the top of the stack (the currently executing frame) on
    down to the root frame.  Optionally, use start to build from an earlier stack
    frame.
    """
    tb = None
    for depth in itertools.count(start + 1, 1):
        try:
            frame = sys._getframe(depth)
            tb = types.TracebackType(tb, frame, frame.f_lasti, frame.f_lineno)
        except ValueError:
            break
    return tb


def try_assert(
    test_func: t.Callable[[], bool],
    fail_msg: str = "",
    timeout_ms: float = 5000,
    attempts: int = 0,
    check_period_ms: int = 20,
):
    tb = create_traceback(start=1)
    timeout_s = abs(timeout_ms) / 1000.0
    check_period_s = abs(check_period_ms) / 1000.0
    if attempts > 0:
        for _attempt_no in range(attempts):
            if test_func():
                return
            time.sleep(check_period_s)
        else:
            att_fail = (
                f"\n  (Still failing after attempt limit [{attempts}], testing every"
                f" {check_period_ms}ms)"
            )
            raise AssertionError(f"{str(fail_msg)}{att_fail}".strip()).with_traceback(
                tb
            )

    elif timeout_s > 0:
        end = time.monotonic() + timeout_s
        while time.monotonic() < end:
            if test_func():
                return
            time.sleep(check_period_s)
        att_fail = (
            f"\n  (Still failing after timeout [{timeout_ms}ms], with attempts "
            f"every {check_period_ms}ms)"
        )
        raise AssertionError(f"{str(fail_msg)}{att_fail}".strip()).with_traceback(tb)

    else:
        raise AssertionError("Bad test configuration: no attempts or timeout period")


def try_for_timeout(
    test_func: t.Callable, timeout_ms: int = 5000, check_period_ms: int = 20
) -> bool:
    timeout_s = abs(timeout_ms) / 1000.0
    check_period_s = abs(check_period_ms) / 1000.0
    end = time.monotonic() + timeout_s
    while time.monotonic() < end:
        if test_func():
            return True
        time.sleep(check_period_s)
    return False


def ez_pack_function(serializer, func, args, kwargs):
    serialized_func = serializer.serialize(func)
    serialized_args = serializer.serialize(args)
    serialized_kwargs = serializer.serialize(kwargs)
    return serializer.pack_buffers(
        [serialized_func, serialized_args, serialized_kwargs]
    )


def double(x: int) -> int:
    return x * 2


def slow_double(x: int, sleep_duration_s: int) -> int:
    import time

    time.sleep(sleep_duration_s)
    return x * 2


def kill_manager():
    import os
    import signal

    manager_pid = os.getppid()
    os.kill(manager_pid, signal.SIGKILL)


def div_zero(x: int):
    return x / 0



from __future__ import annotations

import contextlib
import difflib
import gzip
import json
import logging
import pathlib
import re
import shutil
import sys
import textwrap
import uuid
from datetime import datetime

import click
from click import ClickException
from globus_compute_endpoint.endpoint.config.utils import get_config, load_config_yaml
from globus_compute_endpoint.endpoint.endpoint import Endpoint
from globus_compute_endpoint.exception_handling import handle_auth_errors
from globus_compute_endpoint.logging_config import setup_logging
from globus_compute_endpoint.self_diagnostic import run_self_diagnostic
from globus_compute_endpoint.version import DEPRECATION_FUNCX_ENDPOINT
from globus_compute_sdk.sdk.login_manager import LoginManager
from globus_compute_sdk.sdk.login_manager.tokenstore import ensure_compute_dir
from globus_compute_sdk.sdk.login_manager.whoami import print_whoami_info

try:
    from globus_compute_endpoint.endpoint.endpoint_manager import EndpointManager
except ImportError as _e:
    _has_multi_tenant = False
    if "--debug" in sys.argv and sys.stderr.isatty():
        # We haven't set up logging yet so manually print for now
        print(f"(DEBUG) Failed to import from file: {__file__}", file=sys.stderr)
        print(f"(DEBUG) [{type(_e).__name__}] {_e}", file=sys.stderr)
else:
    _has_multi_tenant = True

log = logging.getLogger(__name__)


class CommandState:
    def __init__(self):
        self.endpoint_config_dir: pathlib.Path = init_config_dir()
        self.debug = False
        self.no_color = False
        self.log_to_console = False
        self.endpoint_uuid = None
        self.die_with_parent = False

    @classmethod
    def ensure(cls) -> CommandState:
        return click.get_current_context().ensure_object(CommandState)


def init_config_dir() -> pathlib.Path:
    try:
        return ensure_compute_dir()
    except (FileExistsError, PermissionError) as e:
        raise ClickException(str(e))


def get_config_dir() -> pathlib.Path:
    state = CommandState.ensure()
    return state.endpoint_config_dir


def get_cli_endpoint() -> Endpoint:
    # this getter creates an Endpoint object from the CommandState
    # it takes its various configurable values from the current CommandState
    # as a result, any number of CLI options may be used to tweak the CommandState
    # via callbacks, and the Endpoint will only be constructed within commands which
    # access the Endpoint via this getter
    state = CommandState.ensure()
    endpoint = Endpoint(debug=state.debug)

    return endpoint


def set_param_to_config(ctx, param, value):
    state = CommandState.ensure()
    setattr(state, param.name, value)
    return state


def log_flag_callback(ctx, param, value):
    if not value or ctx.resilient_parsing:
        return

    state = set_param_to_config(ctx, param, value)
    setup_logging(debug=state.debug, no_color=state.no_color)


def common_options(f):
    f = click.option(
        "--debug",
        is_flag=True,
        expose_value=False,
        callback=log_flag_callback,
        is_eager=True,
        help="Emit extra information generally only helpful for developers",
    )(f)
    f = click.option(
        "--log-to-console",
        is_flag=True,
        expose_value=False,
        is_eager=True,
        callback=log_flag_callback,
        help="Emit log lines to console as well as log",
    )(f)
    f = click.option(
        "--no-color",
        is_flag=True,
        expose_value=False,
        is_eager=True,
        callback=log_flag_callback,
        help="Do not colorize the (console) log lines",
    )(f)

    f = click.help_option("-h", "--help")(f)
    return f


def start_options(f):
    f = click.option(
        "--endpoint-uuid",
        default=None,
        callback=set_param_to_config,
        help="The UUID to register with the Globus Compute services",
    )(f)
    f = click.option(
        "--die-with-parent",
        is_flag=True,
        hidden=True,
        callback=set_param_to_config,
        help="Shutdown if parent process goes away",
    )(f)
    return f


def get_ep_dir_by_name_or_uuid(ctx, param, value):
    conf_dir = get_config_dir()
    try:
        uuid.UUID(value)
    except ValueError:
        # value is a name
        path = conf_dir / value
    else:
        # value is a uuid
        path = Endpoint.get_endpoint_dir_by_uuid(conf_dir, value)
        if path is None:
            msg = textwrap.dedent(
                f"""
                There is no endpoint on this machine with ID '{value}'!

                1. Please create a configuration template with:
                    globus-compute-endpoint configure
                2. Update the configuration
                3. Start the endpoint
                """
            )
            raise ClickException(msg)
    ctx.params["ep_dir"] = path


def verify_not_uuid(ctx, param, value):
    try:
        uuid.UUID(value)
        raise click.BadParameter(
            f"'{ctx.command_path}' requires an endpoint name that"
            f" is not a UUID (got: '{value}')"
        )
    except ValueError:
        return value


def name_or_uuid_arg(f):
    return click.argument(
        "name_or_uuid",
        required=False,
        callback=get_ep_dir_by_name_or_uuid,
        default="default",
        expose_value=False,  # callback supplies ep_dir to command functions
    )(f)


def name_arg(f):
    return click.argument(
        "name", required=False, callback=verify_not_uuid, default="default"
    )(f)


def config_dir_callback(ctx, param, value):
    if value is None or ctx.resilient_parsing:
        return
    state = CommandState.ensure()
    state.endpoint_config_dir = pathlib.Path(value)


@click.group("globus-compute-endpoint")
@click.option(
    "-c",
    "--config-dir",
    default=None,
    help="override default config dir",
    callback=config_dir_callback,
    expose_value=False,
)
def app():
    # the main command group body runs on every command, so the block below will always
    # execute
    setup_logging()  # Until we parse the CLI flags, just setup default logging


@app.command("version")
@common_options
def version_command():
    """Show the version of globus-compute-endpoint"""
    import globus_compute_endpoint as gce

    click.echo(f"Globus Compute endpoint version: {gce.__version__}")


@app.command(name="configure", help="Configure an endpoint")
@click.option("--endpoint-config", default=None, help="a template config to use")
@click.option(
    "--multi-tenant",
    is_flag=True,
    default=False,
    hidden=True,
    help="Configure endpoint as multi-tenant capable",
)
@click.option(
    "--display-name",
    help="A human readable display name for the endpoint, if desired",
)
@name_arg
@common_options
def configure_endpoint(
    *,
    name: str,
    endpoint_config: str | None,
    multi_tenant: bool,
    display_name: str | None,
):
    """Configure an endpoint

    Drops a config.yaml template into the Globus Compute configs directory.
    The template usually goes to ~/.globus_compute/<ENDPOINT_NAME>/config.yaml
    """
    try:
        Endpoint.validate_endpoint_name(name)
    except ValueError as e:
        raise ClickException(str(e))

    if multi_tenant and not _has_multi_tenant:
        raise ClickException("multi-tenant endpoints are not supported on this system")

    compute_dir = get_config_dir()
    ep_dir = compute_dir / name
    Endpoint.configure_endpoint(ep_dir, endpoint_config, multi_tenant, display_name)


@app.command(name="start", help="Start an endpoint")
@name_or_uuid_arg
@start_options
@common_options
@handle_auth_errors
def start_endpoint(*, ep_dir: pathlib.Path, **_kwargs):
    """Start an endpoint

    This function will do:
    1. Connect to the broker service, and register itself
    2. Get connection info from broker service
    3. Start the interchange as a daemon

    |                      Broker service       |
    |               -----2----> Forwarder       |
    |    /register <-----3----+   ^             |
    +-----^-----------------------+-------------+
          |     |                 |
          1     4                 6
          |     v                 |
    +-----+-----+-----+           v
    |      Start      |---5---> EndpointInterchange
    |     Endpoint    |         daemon
    +-----------------+
    """
    state = CommandState.ensure()
    _do_start_endpoint(
        ep_dir=ep_dir,
        endpoint_uuid=state.endpoint_uuid,
        die_with_parent=state.die_with_parent,
    )


@app.command(name="logout", help="Logout from all endpoints")
@click.option(
    "--force",
    is_flag=True,
    help="Revokes tokens even with currently running endpoints",
)
def logout_endpoints(force: bool):
    success, msg = _do_logout_endpoints(force=force)
    if not success:
        # Raising ClickException is apparently the way to do sys.exit(1)
        #     and return a non-zero value to the command line
        # See https://click.palletsprojects.com/en/8.1.x/exceptions/
        if not isinstance(msg, str) or msg is None:
            # Generic unsuccessful if no reason was given
            msg = "Logout unsuccessful"
        raise ClickException(msg)


@app.command("whoami", help="Show the currently logged-in identity")
@click.option(
    "--linked-identities",
    is_flag=True,
    default=False,
    help="Also show identities linked to the currently logged-in primary identity.",
)
def whoami(linked_identities: bool) -> None:
    try:
        print_whoami_info(linked_identities)
    except ValueError as ve:
        raise ClickException(str(ve))


def _do_logout_endpoints(
    force: bool, running_endpoints: dict | None = None
) -> tuple[bool, str | None]:
    """
    Logout from all endpoints and remove cached authentication credentials

    Returns True, None if logout was successful and tokens were found and revoked
    Returns False, error_msg if token revocation was not done
    """
    if running_endpoints is None:
        compute_dir = get_config_dir()
        running_endpoints = Endpoint.get_running_endpoints(compute_dir)
    tokens_revoked = False
    error_msg = None
    if running_endpoints and not force:
        running_list = ", ".join(running_endpoints.keys())
        log.info(
            "The following endpoints are currently running: "
            + running_list
            + "\nPlease use logout --force to proceed"
        )
        error_msg = "Not logging out with running endpoints without --force"
    else:
        tokens_revoked = LoginManager().logout()
        if tokens_revoked:
            log.info("Logout succeeded and all cached credentials were revoked")
        else:
            error_msg = "No cached tokens were found, already logged out?"
            log.info(error_msg)
    return tokens_revoked, error_msg


FUNCX_COMPUTE_IMPORT_UPDATES = {
    "from funcx_endpoint.endpoint.utils.config": "from globus_compute_endpoint.endpoint.config",  # noqa E501
    "from funcx_endpoint.engines": "from globus_compute_endpoint.engines",  # noqa E501
    "from funcx_endpoint.executors": "from globus_compute_endpoint.executors",  # noqa E501
    "from funcx_endpoint.engines": "from globus_compute_endpoint.engines",  # noqa E501
}


def _upgrade_funcx_imports_in_config(ep_dir: pathlib.Path, force=False) -> str:
    """
    This only modifies unindented import lines, as are in the original
    config.py.  Indented matching lines are user created and would have to be
    updated manually by the user.

    The force flag will overwrite an existing (non-directory) config.py.bak

    This method uses a randomly generated intermediate output file in case
    there are any permission or unforeseen file system issues
    """
    name = ep_dir.name
    config_path = ep_dir / "config.py"
    config_backup = ep_dir / "config.py.bak"

    try:
        config_text = config_path.read_text()
        upd_config_text = config_text

        for pattern, repl in FUNCX_COMPUTE_IMPORT_UPDATES.items():
            upd_config_text = re.sub(f"^{pattern}", repl, upd_config_text, flags=re.M)

        if upd_config_text == config_text:
            return f"No funcX import statements found in config.py for {name}"

        change_diff = "".join(
            difflib.unified_diff(
                config_text.splitlines(keepends=True),
                upd_config_text.splitlines(keepends=True),
                n=3,  # Typical 3 lines of context
            )
        )

        if config_backup.exists() and not force:
            msg = (
                f"{config_backup} already exists.\n"
                "Rename it or use the --force flag to update config."
            )
            raise ClickException(msg)
        elif config_backup.is_dir():
            msg = (
                f"{config_backup} is a directory.\n"
                "Rename it before proceeding with config update."
            )
            raise ClickException(msg)

        # Write to temporary file in case of issues
        tmp_output_path = ep_dir / ("config.py." + uuid.uuid4().hex)
        tmp_output_path.write_text(upd_config_text)

        # Rename files last, as it's the least likely to err
        if sys.version_info < (3, 8):
            try:
                # Dropping support for Python 3.7 "real soon", so this branch will
                # soon go away
                config_backup.unlink()
            except FileNotFoundError:
                pass
        else:
            config_backup.unlink(missing_ok=True)
        shutil.move(config_path, config_backup)  # Preserve file timestamp
        shutil.move(tmp_output_path, config_path)

        return (
            f"Applied following diff for endpoint {name}:\n{change_diff}\n\n"
            f"  The previous config has been renamed to {config_backup}"
        )

    except FileNotFoundError as err:
        msg = f"No config.py was found for endpoint ({name}) in {ep_dir}"
        raise ClickException(msg) from err
    except ClickException:
        raise
    except Exception as err:
        msg = f"Unknown Exception {err} attempting to reformat config.py in {ep_dir}"
        raise ClickException(msg) from err


def _do_start_endpoint(
    *,
    ep_dir: pathlib.Path,
    endpoint_uuid: str | None,
    die_with_parent: bool = False,
):
    state = CommandState.ensure()
    if ep_dir.is_dir():
        setup_logging(
            logfile=ep_dir / "endpoint.log",
            debug=state.debug,
            console_enabled=state.log_to_console,
            no_color=state.no_color,
        )

    reg_info = {}
    config_str = None
    if sys.stdin and not (sys.stdin.closed or sys.stdin.isatty()):
        try:
            stdin_data = json.loads(sys.stdin.read())

            if not isinstance(stdin_data, dict):
                type_name = stdin_data.__class__.__name__
                raise ValueError(
                    "Expecting JSON dictionary with endpoint info; got"
                    f" {type_name} instead"
                )

            reg_info = stdin_data.get("amqp_creds", {})
            config_str = stdin_data.get("config", None)

        except Exception as e:
            exc_type = e.__class__.__name__
            log.debug("Invalid info on stdin -- (%s) %s", exc_type, e)

    if config_str is not None:
        ep_config = load_config_yaml(config_str)
    else:
        ep_config = get_config(ep_dir)

    if die_with_parent:
        # The endpoint cannot die with its parent if it
        # doesn't have one :)
        ep_config.detach_endpoint = False
        log.debug("The --die-with-parent flag has set detach_endpoint to False")

    if ep_config.multi_tenant:
        if not _has_multi_tenant:
            raise ClickException(
                "multi-tenant endpoints are not supported on this system"
            )
        epm = EndpointManager(ep_dir, endpoint_uuid, ep_config, reg_info)
        epm.start()
    else:
        get_cli_endpoint().start_endpoint(
            ep_dir,
            endpoint_uuid,
            ep_config,
            state.log_to_console,
            state.no_color,
            reg_info,
            die_with_parent,
        )


@app.command("stop")
@name_or_uuid_arg
@click.option(
    "--remote",
    is_flag=True,
    help="send stop signal to all endpoints with this UUID, local or elsewhere",
)
@common_options
@handle_auth_errors
def stop_endpoint(*, ep_dir: pathlib.Path, remote: bool):
    """Stops an endpoint using the pidfile"""
    _do_stop_endpoint(ep_dir=ep_dir, remote=remote)


@app.command("update_funcx_config")
@name_or_uuid_arg
@click.option(
    "--force",
    is_flag=True,
    default=False,
    help="update config and backup to config.py.bak even if it already exists",
)
def update_funcx_endpoint_config(*, ep_dir: pathlib.Path, force: bool):
    """
    Update imports file from funcx_endpoint.* to globus_compute_endpoint.*

    Either should raise ClickException or returns modification result message
    """
    print(_upgrade_funcx_imports_in_config(ep_dir=ep_dir, force=force))


def _do_stop_endpoint(*, ep_dir: pathlib.Path, remote: bool = False) -> None:
    Endpoint.stop_endpoint(ep_dir, get_config(ep_dir), remote=remote)


@app.command("restart")
@name_or_uuid_arg
@common_options
@start_options
def restart_endpoint(*, ep_dir: pathlib.Path, **_kwargs):
    """Restarts an endpoint"""
    state = CommandState.ensure()
    _do_stop_endpoint(ep_dir=ep_dir)
    _do_start_endpoint(
        ep_dir=ep_dir,
        endpoint_uuid=state.endpoint_uuid,
        die_with_parent=state.die_with_parent,
    )


@app.command("list")
@common_options
def list_endpoints():
    """List all available endpoints"""
    compute_dir = get_config_dir()
    Endpoint.print_endpoint_table(compute_dir)


@app.command("delete")
@name_or_uuid_arg
@click.option(
    "--force",
    default=False,
    is_flag=True,
    help="Deletes the local endpoint even if the web service returns an error.",
)
@click.option(
    "--yes", default=False, is_flag=True, help="Do not ask for confirmation to delete."
)
@common_options
@handle_auth_errors
def delete_endpoint(*, ep_dir: pathlib.Path, force: bool, yes: bool):
    """Deletes an endpoint and its config."""
    if not yes:
        click.confirm(
            f"Are you sure you want to delete the endpoint named <{ep_dir.name}>?",
            abort=True,
        )

    Endpoint.delete_endpoint(ep_dir, get_config(ep_dir), force=force)


@app.command("self-diagnostic")
@click.option(
    "-z",
    "--gzip",
    "compress",
    default=False,
    is_flag=True,
    help="Save the output to a Gzip-compressed file.",
)
@click.option(
    "--log-kb",
    default=5120,
    help=(
        "Specify the number of kilobytes (KB) to read from log files."
        " Defaults to 5,120 KB (5 MB)."
    ),
)
@click.help_option("-h", "--help")
def self_diagnostic(compress: bool, log_kb: int):
    """Run several diagnostic commands to help identify issues.

    This may produce a large amount of output.
    """
    log_bytes = log_kb * 1024

    if not compress:
        run_self_diagnostic(log_bytes=log_bytes)
    else:
        current_date = datetime.now().strftime("%Y-%m-%d")
        filename = f"globus_compute_diagnostic_{current_date}.txt.gz"

        with gzip.open(filename, "wb") as f:
            with contextlib.redirect_stdout(f):  # type: ignore[type-var]
                run_self_diagnostic(log_bytes=log_bytes)

        click.echo(f"Successfully created {filename}")


def cli_run():
    """Entry point for setuptools to point to"""
    app()


def cli_run_funcx():
    """Entry point that prints a custom message. i.e. deprecation warnings"""
    fmt = DEPRECATION_FUNCX_ENDPOINT

    # Colorized notice to be a bit more visible
    fmt = "{title}DEPRECATION NOTICE{rs}\n{body}" + fmt + "{rs}"
    title = frs = body = rs = ""
    if sys.stderr.isatty():
        title = "\033[37;41m"  # White FG, Red BG
        body = "\033[33m"  # Yellow FG
        rs = "\033[0m"  # Reset colors

    print(fmt.format(title=title, body=body, frs=frs, rs=rs), file=sys.stderr)
    app()


if __name__ == "__main__":
    app()



import logging
import os

from globus_compute_endpoint.engines.high_throughput.worker_map import WorkerMap


class TestWorkerMap:
    def test_add_worker(self, mocker):
        mock_popen = mocker.patch(
            "globus_compute_endpoint.engines.high_throughput.worker_map.subprocess.Popen"  # noqa: E501
        )
        mock_popen.return_value = "proc"

        # Test adding with no accelerators
        worker_map = WorkerMap(1, [])
        worker = worker_map.add_worker(
            worker_id="0",
            address="127.0.0.1",
            debug=logging.DEBUG,
            uid="test1",
            logdir=os.getcwd(),
            worker_port=50001,
        )

        assert list(worker.keys()) == ["0"]
        assert worker["0"] == "proc"
        assert worker_map.worker_id_counter == 1
        assert worker_map.available_accelerators is None

        # Test with an accelerator
        worker_map = WorkerMap(1, ["0"])
        worker_map.add_worker(
            worker_id="1",
            address="127.0.0.1",
            debug=logging.DEBUG,
            uid="test1",
            logdir=os.getcwd(),
            worker_port=50001,
        )

        last_call = mock_popen.mock_calls[-1]
        assert last_call[-1]["env"]["CUDA_VISIBLE_DEVICES"] == "0"



from __future__ import annotations

import random
import signal
import string
import time
import typing as t
import uuid
from queue import Queue

import globus_compute_sdk as gc
import globus_sdk
import pytest
import responses
from globus_compute_endpoint import engines
from globus_compute_endpoint.engines.base import GlobusComputeEngineBase
from globus_compute_sdk.sdk.web_client import WebClient


@pytest.fixture(autouse=True)
def verify_all_tests_reset_signals():
    orig_sig_handlers = [
        (sig, signal.getsignal(sig))
        for sig in range(1, 21)
        if sig not in (signal.SIGKILL, signal.SIGSTOP)  # the uncatchables
    ]
    yield
    handlers = [(sig, sigh, signal.getsignal(sig)) for sig, sigh in orig_sig_handlers]
    for sig, bef, aft in handlers:
        assert bef == aft, f"Signal {sig} not reset"


@pytest.fixture
def reset_signals():
    # As identified by an interaction with Parsl during our UTs, ensure that we
    # return the signals after each test.  Interprocess play is fun ...
    orig_sig_handlers = [
        (sig, signal.getsignal(sig))
        for sig in range(1, 21)
        if sig not in (signal.SIGKILL, signal.SIGSTOP)  # the uncatchables
    ]

    yield
    for sig, sigh in orig_sig_handlers:
        signal.signal(sig, sigh)


@pytest.fixture(scope="session")
def endpoint_uuid():
    return str(uuid.UUID(int=0))


@pytest.fixture(scope="session")
def default_endpoint_id():
    return str(uuid.UUID(int=1))


@pytest.fixture(scope="session")
def other_endpoint_id():
    return str(uuid.UUID(int=2))


@pytest.fixture(scope="session")
def tod_session_num():
    yield round(time.time()) % 86400


class FakeLoginManager:
    def ensure_logged_in(self) -> None:
        ...

    def logout(self) -> bool:
        ...

    def get_auth_client(self) -> globus_sdk.AuthClient:
        return globus_sdk.AuthClient(authorizer=globus_sdk.NullAuthorizer())

    def get_web_client(self, *, base_url: str | None = None) -> WebClient:
        return WebClient(
            base_url="https://compute.api.globus.org/v2/",
            authorizer=globus_sdk.NullAuthorizer(),
        )


@pytest.fixture
def get_standard_compute_client():
    responses.add(
        method=responses.GET,
        url="https://compute.api.globus.org/v2/version",
        headers={"Content-Type": "application/json"},
        json={"api": "0.4.0", "min_ep_version": "0.0.0", "min_sdk_version": "0.0.0"},
    )

    def func():
        return gc.Client(
            login_manager=FakeLoginManager(),
            do_version_check=False,
        )

    return func


###
# Engines
###


@pytest.fixture
def engine_heartbeat() -> float:
    return 0.1


@pytest.fixture
def engine_runner(tmp_path, engine_heartbeat) -> t.Callable:
    engines_to_shutdown = []

    def _runner(engine_type: t.Type[GlobusComputeEngineBase]):
        ep_id = uuid.uuid4()
        queue = Queue()
        if engine_type is engines.ProcessPoolEngine:
            k = dict(heartbeat_period_s=engine_heartbeat, max_workers=2)
        elif engine_type is engines.ThreadPoolEngine:
            k = dict(heartbeat_period_s=engine_heartbeat, max_workers=2)
        elif engine_type is engines.GlobusComputeEngine:
            k = dict(
                address="127.0.0.1",
                heartbeat_period_s=engine_heartbeat,
                heartbeat_threshold=1,
            )
        else:
            raise NotImplementedError(f"Unimplemented: {engine_type.__name__}")
        engine = engine_type(**k)
        engine.start(
            endpoint_id=ep_id, run_dir=str(tmp_path), results_passthrough=queue
        )
        engines_to_shutdown.append(engine)
        return engine

    yield _runner
    for ngin in engines_to_shutdown:
        ngin.shutdown(block=True)


###
# Misc
###


@pytest.fixture
def randomstring():
    def func(length=5, alphabet=string.ascii_letters):
        return "".join(random.choice(alphabet) for _ in range(length))

    return func


@pytest.fixture
def noop():
    def _wrapped(*args, **kwargs):
        pass

    return _wrapped



from __future__ import annotations

import logging
import os
import random
import subprocess
import time
from collections import defaultdict
from queue import Empty, Queue
from typing import Any

from globus_compute_endpoint.logging_config import ComputeLogger

log: ComputeLogger = logging.getLogger(__name__)  # type: ignore


class WorkerMap:
    """WorkerMap keeps track of workers"""

    def __init__(
        self,
        max_worker_count: int,
        available_accelerators: list[str],
    ):
        """

        Parameters
        ----------
        max_worker_count:
            Maximum number of workers allowed
        available_accelerators:
            List of accelerator devices workers can be pinned to
        """
        self.max_worker_count = max_worker_count
        self.total_worker_type_counts: dict[str, int] = {
            "unused": self.max_worker_count
        }
        self.ready_worker_type_counts: dict[str, int] = {
            "unused": self.max_worker_count
        }
        self.pending_worker_type_counts: dict[str, Any] = {}
        # a dict to keep track of all the worker_queues with the key of work_type
        self.worker_queues: dict[str, Any] = {}
        # a dict to keep track of all the worker_types with the key of worker_id
        self.worker_types: dict[str, str] = {}
        self.worker_id_counter = 0  # used to create worker_ids

        # Only spin up containers if active_workers + pending_workers < max_workers.
        self.active_workers = 0
        self.pending_workers = 0

        # Need to keep track of workers that are ABOUT to die
        self.to_die_count: dict[str, int] = {}

        # Need to keep track of workers' last idle time by worker type
        self.worker_idle_since: dict[str, float] = {}

        # Create a queue of available accelerators, if accelerators are defined
        self.available_accelerators: Queue | None = None
        if len(available_accelerators) != 0:
            self.available_accelerators = Queue()
            for device in available_accelerators:
                self.available_accelerators.put(device)
        self.assigned_accelerators: dict[str, str] = {}  # Map worker ID -> accelerator

        self._noisy_log: dict[str, Any] = defaultdict(dict)

    def register_worker(self, worker_id, worker_type):
        """Add a new worker"""
        log.debug(f"In register worker worker_id: {worker_id} type:{worker_type}")
        self.worker_types[worker_id] = worker_type

        if worker_type not in self.worker_queues:
            self.worker_queues[worker_type] = Queue()

        self.total_worker_type_counts[worker_type] = (
            self.total_worker_type_counts.get(worker_type, 0) + 1
        )
        self.ready_worker_type_counts[worker_type] = (
            self.ready_worker_type_counts.get(worker_type, 0) + 1
        )
        self.pending_worker_type_counts[worker_type] = (
            self.pending_worker_type_counts.get(worker_type, 0) - 1
        )
        self.pending_workers -= 1
        self.active_workers += 1
        self.worker_queues[worker_type].put(worker_id)
        self.worker_idle_since[worker_type] = time.time()

        if worker_type not in self.to_die_count:
            self.to_die_count[worker_type] = 0

    def start_remove_worker(self, worker_type):
        """Increase the to_die_count in prep for a worker getting removed"""
        self.to_die_count[worker_type] += 1

    def remove_worker(self, worker_id):
        """Remove the worker from the WorkerMap

        Should already be KILLed by this point.
        """
        worker_type = self.worker_types[worker_id]
        self.active_workers -= 1
        self.total_worker_type_counts[worker_type] -= 1
        self.to_die_count[worker_type] -= 1
        self.total_worker_type_counts["unused"] += 1
        self.ready_worker_type_counts["unused"] += 1

        # Mark the accelerator as available, if provided
        if worker_id in self.assigned_accelerators:
            device = self.assigned_accelerators.pop(worker_id)
            self.available_accelerators.put(device)

    def spin_up_workers(
        self,
        next_worker_q,
        mode="no_container",
        container_cmd_options="",
        address=None,
        debug=None,
        uid=None,
        logdir=None,
        worker_port=None,
    ):
        """Helper function to call 'remove' for appropriate workers in 'new_worker_map'.

        Parameters
        ----------
        new_worker_q : queue.Queue()
            Queue of worker types to be spun up next.
        mode : str
            Mode of the worker, no_container, singularity, etc.
        address : str
            Address at which to connect to the workers.
        debug : bool
            Whether debug logging is activated.
        uid : str
            Worker ID to be assigned to worker.
        logdir: str
            Directory in which to write logs
        worker_port: int
            Port at which to connect to the workers.

        Returns
        ---------
        Total number of spun-up workers.
        """
        spin_ups = {}

        log.trace("Next Worker Qsize: %s", len(next_worker_q))
        log.trace("Active Workers: %s", self.active_workers)
        log.trace("Pending Workers: %s", self.pending_workers)
        log.trace("Max Worker Count: %s", self.max_worker_count)

        if (
            len(next_worker_q) > 0
            and self.active_workers + self.pending_workers < self.max_worker_count
        ):
            log.debug("Spinning up new workers")
            log.debug(
                "Empty slots: %s",
                self.max_worker_count - self.active_workers - self.pending_workers,
            )
            log.debug(f"New workers: {len(next_worker_q)}")
            log.debug(f"Unused slots: {self.total_worker_type_counts['unused']}")
            num_slots = min(
                self.max_worker_count - self.active_workers - self.pending_workers,
                len(next_worker_q),
                self.total_worker_type_counts["unused"],
            )
            for _ in range(num_slots):
                try:
                    proc = self.add_worker(
                        worker_id=str(self.worker_id_counter),
                        worker_type=next_worker_q.pop(0),
                        container_cmd_options=container_cmd_options,
                        mode=mode,
                        address=address,
                        debug=debug,
                        uid=uid,
                        logdir=logdir,
                        worker_port=worker_port,
                    )
                except Exception:
                    log.exception("Error spinning up worker! Skipping...")
                    continue
                else:
                    spin_ups.update(proc)
        return spin_ups

    def spin_down_workers(
        self,
        new_worker_map,
        worker_max_idletime=60,
        need_more=False,
        scheduler_mode="hard",
    ):
        """Helper function to call 'remove' for appropriate workers in 'new_worker_map'.

        Parameters
        ----------
        new_worker_map : dict
           {worker_type: total_number_of_containers,...}.
        need_more: bool
            whether the manager needs to spin down some warm containers
        Returns
        ---------
        List of removed worker types.
        """
        if need_more:
            return self._spin_down(
                new_worker_map,
                worker_max_idletime=worker_max_idletime,
                scheduler_mode=scheduler_mode,
                check_idle=False,
            )
        else:
            return self._spin_down(
                new_worker_map,
                worker_max_idletime=worker_max_idletime,
                scheduler_mode=scheduler_mode,
                check_idle=True,
            )

    def _spin_down(
        self,
        new_worker_map,
        worker_max_idletime=60,
        scheduler_mode="hard",
        check_idle=True,
    ):
        """Helper function to call 'remove' for appropriate workers in 'new_worker_map'.

        Parameters
        ----------
        new_worker_map : dict
           {worker_type: total_number_of_containers,...}.
        check_idle : boolean
           A boolean to indicate whether to check the idle time of containers or not

           If checked, that means the workloads are not so busy, and we can leave the
           container workers alive until the worker_max_idletime is reached.  Otherwise,
           that means the workloads are busy and we need to turn of some containers to
           acommodate the workers, regardless of if it reaches the worker_max_idletime.

        Returns
        ---------
        List of removed worker types.
        """
        spin_downs = []
        container_switch_count = 0
        now = time.time()
        for worker_type in self.total_worker_type_counts:
            if worker_type == "unused":
                continue
            if (
                check_idle
                and now - self.worker_idle_since[worker_type] < worker_max_idletime
            ):
                log.trace(
                    "Current time: %s (idle since: %s).  Worker type %s has not "
                    "exceeded maximum idle time %s; continuing",
                    now,
                    self.worker_idle_since[worker_type],
                    worker_type,
                    worker_max_idletime,
                )
                continue
            num_remove = max(
                0,
                self.total_worker_type_counts[worker_type]
                - self.to_die_count.get(worker_type, 0)
                - new_worker_map.get(worker_type, 0),
            )
            if scheduler_mode == "hard":
                # Leave at least one worker alive in hard mode
                max_remove = max(0, self.total_worker_type_counts[worker_type] - 1)
                num_remove = min(num_remove, max_remove)

            if num_remove > 0:
                log.debug(f"Removing {num_remove} workers of type {worker_type}")
            for _i in range(num_remove):
                spin_downs.append(worker_type)
            # A container switching is defined as a warm container must be
            # switched to another container to accommodate the workloads.
            # If a container worker has been idle for worker_max_idletime,
            # Then it is not counted as a container switching
            if not check_idle:
                container_switch_count += num_remove
        return spin_downs, container_switch_count

    def add_worker(
        self,
        worker_id=None,
        mode="no_container",
        worker_type="RAW",
        container_cmd_options="",
        walltime=1,
        address=None,
        debug=None,
        worker_port=None,
        logdir=None,
        uid=None,
    ):
        """Launch the appropriate worker

        Parameters
        ----------
        worker_id : str
           Worker identifier string
        mode : str
           Valid options are no_container, singularity
        walltime : int
           Walltime in seconds before we check status

        """
        if worker_id is None:
            str(random.random())

        debug = " --debug" if debug else ""

        worker_id = f" --worker_id {worker_id}"

        self.worker_id_counter += 1

        cmd = (
            f"globus-compute-worker {debug}{worker_id} "
            f"-a {address} "
            f"-p {worker_port} "
            f"-t {worker_type} "
            f"--logdir={os.path.join(logdir, uid)} "
        )

        container_uri = None
        if worker_type != "RAW":
            container_uri = worker_type

        # If accelerator list is provided, get the next one off the queue
        #   and mark it as assigned
        environment_variables = os.environ.copy()
        if self.available_accelerators is not None:
            try:
                device = self.available_accelerators.get_nowait()
            except Empty:
                raise ValueError(
                    "No accelerators are available."
                    " New worker must be created only"
                    " after another is removed"
                )
            self.assigned_accelerators[worker_id] = device
            log.info(f"Assigned worker '{worker_id}' to accelerator '{device}'")

            # Create the
            #  TODO (wardlt): This code has only been tested for CUDA
            environment_variables["CUDA_VISIBLE_DEVICES"] = device
            environment_variables["ROCR_VISIBLE_DEVICES"] = device
            environment_variables["SYCL_DEVICE_FILTER"] = f"*:*:{device}"

        log.info(f"Command string :\n {cmd}")
        log.info(f"Mode: {mode}")
        log.info(f"Container uri: {container_uri}")
        log.info(f"Container cmd options: {container_cmd_options}")
        log.info(f"Worker type: {worker_type}")

        if mode == "no_container":
            modded_cmd = cmd
        elif mode == "singularity_reuse":
            if container_uri is None:
                log.warning(
                    "No container is specified for singularity mode. "
                    "Spawning a worker in a raw process instead."
                )
                modded_cmd = cmd
            elif not os.path.exists(container_uri):
                log.warning(
                    f"Container uri {container_uri} is not found. "
                    "Spawning a worker in a raw process instead."
                )
                modded_cmd = cmd
            else:
                modded_cmd = (
                    f"singularity exec {container_cmd_options} {container_uri} {cmd}"
                )
            log.info(f"Command string with singularity:\n {modded_cmd}")
        else:
            raise NameError("Invalid container launch mode.")

        try:
            proc = subprocess.Popen(
                modded_cmd.split(),
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
                shell=False,
                env=environment_variables,
            )

        except Exception:
            log.exception("Got an error in worker launch")
            raise

        self.total_worker_type_counts["unused"] -= 1
        self.ready_worker_type_counts["unused"] -= 1
        self.pending_worker_type_counts[worker_type] = (
            self.pending_worker_type_counts.get(worker_type, 0) + 1
        )
        self.pending_workers += 1

        return {str(self.worker_id_counter - 1): proc}

    def get_next_worker_q(self, new_worker_map) -> tuple[list[str], bool]:
        """Helper function to generate a queue of next workers to spin up .
            From a mapping generated by the scheduler

        Parameters
        ----------
        new_worker_map : dict
           {worker_type: total_number_of_containers,...}

        Returns
        ---------
        Queue containing the next workers the system should spin-up.
        """

        _log_data = self._noisy_log["get_next_worker_q"]
        _l_total = _log_data.get("total_worker_type_counts")
        _l_pending = _log_data.get("pending_worker_type_counts")
        if (_l_total, _l_pending) != (
            self.total_worker_type_counts,
            self.pending_worker_type_counts,
        ):
            log.debug(
                "total_worker_type_counts: %s; pending_worker_type_counts: %s",
                self.total_worker_type_counts,
                self.pending_worker_type_counts,
            )
            _log_data["total_worker_type_counts"] = self.total_worker_type_counts
            _log_data["pending_worker_type_counts"] = self.pending_worker_type_counts

        new_worker_list = []
        for worker_type in new_worker_map:
            cur_workers = self.total_worker_type_counts.get(
                worker_type, 0
            ) + self.pending_worker_type_counts.get(worker_type, 0)
            if new_worker_map[worker_type] > cur_workers:
                for _i in range(new_worker_map[worker_type] - cur_workers):
                    # Add worker
                    new_worker_list.append(worker_type)

        # need_more is to reflect if a manager needs more workers than the current
        # unused slots
        # If yes, that means the manager needs to turn off some warm workers to serve
        # the requests
        need_more = len(new_worker_list) > self.total_worker_type_counts["unused"]
        # Randomly assign order of newly needed containers... add to spin-up queue.
        if len(new_worker_list) > 0:
            random.shuffle(new_worker_list)

        return new_worker_list, need_more

    def update_worker_idle(self, worker_type):
        """Update the workers' last idle time by worker type"""
        log.debug(f"Worker idle since: {self.worker_idle_since}")
        self.worker_idle_since[worker_type] = time.time()

    def put_worker(self, worker):
        """Adds worker to the list of waiting workers"""
        worker_type = self.worker_types[worker]

        if worker_type not in self.worker_queues:
            self.worker_queues[worker_type] = Queue()

        self.ready_worker_type_counts[worker_type] += 1
        self.worker_queues[worker_type].put(worker)

    def get_worker(self, worker_type):
        """Get a task and reduce the # of worker for that type by 1.
        Raises queue.Empty if empty
        """
        worker = self.worker_queues[worker_type].get_nowait()
        self.ready_worker_type_counts[worker_type] -= 1
        return worker

    def get_worker_counts(self):
        """Returns just the dict of worker_type and counts"""
        return self.total_worker_type_counts

    def ready_worker_count(self):
        return sum(self.ready_worker_type_counts.values())

    def advertisement(self):
        """
        Manager capacity advertisement to interchange.

        The advertisement includes two parts:

        One is the read_worker_type_counts, which reflects the capacity of different
        types of containers on the manager.

        The other is the total number of workers of each type. This includes all the
        pending workers and to_die workers when advertising. We need this "total"
        advertisement because we use killer task mechanisms to kill a worker. When a
        manager is advertising, there may be some killer tasks in queue, and we want to
        ensure that the manager does not over-advertise its actual capacity. Instead,
        let the interchange decide if it is sending too many tasks to the manager.
        """
        ads = {"total": {}, "free": {}}
        total = dict(self.total_worker_type_counts)
        for worker_type in self.pending_worker_type_counts:
            total[worker_type] = (
                total.get(worker_type, 0)
                + self.pending_worker_type_counts[worker_type]
                - self.to_die_count.get(worker_type, 0)
            )
        ads["total"].update(total)
        ads["free"].update(self.ready_worker_type_counts)
        return ads



from __future__ import annotations

import fcntl
import json
import logging
import os
import pathlib
import pwd
import queue
import re
import resource
import signal
import socket
import sys
import threading
import time
import typing as t
from datetime import datetime

import globus_compute_sdk as GC

try:
    import pyprctl
except AttributeError as e:
    raise ImportError("pyprctl is not supported on this system") from e
import setproctitle
import yaml
from globus_compute_endpoint import __version__
from globus_compute_endpoint.endpoint.config import Config
from globus_compute_endpoint.endpoint.config.utils import (
    load_user_config_schema,
    render_config_user_template,
    serialize_config,
)
from globus_compute_endpoint.endpoint.endpoint import Endpoint
from globus_compute_endpoint.endpoint.rabbit_mq.command_queue_subscriber import (
    CommandQueueSubscriber,
)
from globus_compute_endpoint.endpoint.utils import _redact_url_creds, is_privileged
from globus_sdk import GlobusAPIError, NetworkError

if t.TYPE_CHECKING:
    from pika.spec import BasicProperties


log = logging.getLogger(__name__)


class InvalidCommandError(Exception):
    pass


class InvalidUserError(Exception):
    pass


class EndpointManager:
    def __init__(
        self,
        conf_dir: pathlib.Path,
        endpoint_uuid: str | None,
        config: Config,
        reg_info: dict | None = None,
    ):
        log.info("Endpoint Manager initialization")

        self.conf_dir = conf_dir
        self._reload_requested = False
        self._time_to_stop = False
        self._kill_event = threading.Event()

        self._child_args: dict[int, tuple[int, int, str, str]] = {}
        self._wait_for_child = False

        self._command_queue: queue.SimpleQueue[
            tuple[int, BasicProperties, bytes]
        ] = queue.SimpleQueue()
        self._command_stop_event = threading.Event()

        endpoint_uuid = Endpoint.get_or_create_endpoint_uuid(conf_dir, endpoint_uuid)

        if not reg_info:
            try:
                client_options = {
                    "funcx_service_address": config.funcx_service_address,
                    "environment": config.environment,
                }

                gcc = GC.Client(**client_options)
                reg_info = gcc.register_endpoint(
                    conf_dir.name,
                    endpoint_uuid,
                    metadata=EndpointManager.get_metadata(config, conf_dir),
                    multi_tenant=True,
                )

                # Mostly to appease mypy, but also a useful text if it ever
                # *does* happen
                assert reg_info is not None, "Empty response from Compute API"

            except GlobusAPIError as e:
                if e.http_status == 409 or e.http_status == 423:
                    # RESOURCE_CONFLICT or RESOURCE_LOCKED
                    blocked_msg = f"Endpoint registration blocked.  [{e.text}]"
                    print(blocked_msg)
                    log.warning(blocked_msg)
                    exit(os.EX_UNAVAILABLE)
                raise
            except NetworkError as e:
                log.exception("Network error while registering multi-tenant endpoint")
                log.critical(f"Network failure; unable to register endpoint: {e}")
                exit(os.EX_TEMPFAIL)

        upstream_ep_uuid = reg_info.get("endpoint_id")
        if upstream_ep_uuid != endpoint_uuid:
            log.error(
                "Unexpected response from server: mismatched endpoint id."
                f"\n  Expected: {endpoint_uuid}, received: {upstream_ep_uuid}"
            )
            exit(os.EX_SOFTWARE)

        self._endpoint_uuid_str = upstream_ep_uuid

        try:
            cq_info = reg_info["command_queue_info"]
            _ = cq_info["connection_url"], cq_info["queue"]
        except Exception as e:
            log.debug("%s", reg_info)
            log.error(
                "Invalid or unexpected registration data structure:"
                f" ({e.__class__.__name__}) {e}"
            )
            exit(os.EX_DATAERR)

        self._mt_user = pwd.getpwuid(os.getuid())
        if config.force_mt_allow_same_user:
            self._allow_same_user = True
            log.warning(
                "Configuration item `force_mt_allow_same_user` set to true; this is"
                " considered a very dangerous override -- please use with care,"
                " especially if allowing this endpoint to be utilized by multiple"
                " users."
                f"\n  Endpoint (UID, GID): ({os.getuid()}, {os.getgid()}) "
            )
        else:
            self._allow_same_user = not is_privileged(self._mt_user)

        # sanitize passwords in logs
        log_reg_info = re.subn(r"://.*?@", r"://***:***@", repr(reg_info))
        log.debug(f"Registration information: {log_reg_info}")

        json_file = conf_dir / "endpoint.json"

        # `endpoint_id` key kept for backward compatibility when
        # globus-compute-endpoint list is called
        ep_info = {"endpoint_id": endpoint_uuid}
        json_file.write_text(json.dumps(ep_info))
        log.debug(f"Registration info written to {json_file}")

        # * == "multi-tenant"; not important until it is, so let it be subtle
        ptitle = f"Globus Compute Endpoint *({endpoint_uuid}, {conf_dir.name})"
        if config.environment:
            ptitle += f" - {config.environment}"
        ptitle += f" [{setproctitle.getproctitle()}]"
        setproctitle.setproctitle(ptitle)

        self._command = CommandQueueSubscriber(
            queue_info=cq_info,
            command_queue=self._command_queue,
            stop_event=self._command_stop_event,
            thread_name="CQS",
        )

    @staticmethod
    def get_metadata(config: Config, conf_dir: pathlib.Path) -> dict:
        # Piecemeal Config settings because for MT, most of the ST items are
        # unrelated -- the MT (aka EndpointManager) does not execute tasks
        return {
            "endpoint_version": __version__,
            "hostname": socket.getfqdn(),
            "local_user": pwd.getpwuid(os.getuid()).pw_name,
            "config": serialize_config(config),
            "user_config_schema": load_user_config_schema(conf_dir),
        }

    def request_shutdown(self, sig_num, curr_stack_frame):
        self._time_to_stop = True

    def set_child_died(self, sig_num, curr_stack_fframe):
        self._wait_for_child = True

    def wait_for_children(self):
        try:
            self._wait_for_child = False
            wait_flags = os.WNOHANG
            pid, exit_status_ind = os.waitpid(-1, wait_flags)
            while pid > 0:
                try:
                    rc = os.waitstatus_to_exitcode(exit_status_ind)
                except ValueError:
                    rc = -127  # invalid signal number

                *_, proc_args = self._child_args.pop(pid, (None, None, None, None))
                proc_args = f" [{proc_args}]" if proc_args else ""
                if not rc:
                    log.info(f"Command stopped normally ({pid}){proc_args}")
                elif rc > 0:
                    log.warning(f"Command return code: {rc} ({pid}){proc_args}")
                elif rc == -127:
                    log.warning(f"Command unknown return code: ({pid}){proc_args}")
                else:
                    log.warning(
                        f"Command terminated by signal: {-rc} ({pid}){proc_args}"
                    )
                pid, exit_status_ind = os.waitpid(-1, wait_flags)

        except ChildProcessError:
            pass
        except Exception as e:
            log.exception(f"Failed to wait for a child process: {e}")

    def _install_signal_handlers(self):
        signal.signal(signal.SIGTERM, self.request_shutdown)
        signal.signal(signal.SIGINT, self.request_shutdown)
        signal.signal(signal.SIGQUIT, self.request_shutdown)

        signal.signal(signal.SIGCHLD, self.set_child_died)

    def start(self):
        log.info(f"\n\n========== Endpoint Manager begins: {self._endpoint_uuid_str}")

        msg_out = None
        if sys.stdout.isatty():
            msg_out = sys.stdout
        elif sys.stderr.isatty():
            msg_out = sys.stderr

        if msg_out:
            hl, r = "\033[104m", "\033[m"
            pld = f"{hl}{self._endpoint_uuid_str}{r}"
            print(f"        >>> Multi-Tenant Endpoint ID: {pld} <<<", file=msg_out)

        self._install_signal_handlers()

        try:
            self._event_loop()
        except Exception:
            log.exception("Unhandled exception; shutting down endpoint master")

        ptitle = f"[shutdown in progress] {setproctitle.getproctitle()}"
        setproctitle.setproctitle(ptitle)
        self._command_stop_event.set()
        self._kill_event.set()

        os.killpg(os.getpgid(0), signal.SIGTERM)

        proc_uid, proc_gid = os.getuid(), os.getgid()
        for msg_prefix, signum in (
            ("Signaling shutdown", signal.SIGTERM),
            ("Forcibly killing", signal.SIGKILL),
        ):
            for pid, (uid, gid, uname, proc_args) in list(self._child_args.items()):
                proc_ident = f"PID: {pid}, UID: {uid}, GID: {gid}, User: {uname}"
                log.info(f"{msg_prefix} of user endpoint ({proc_ident}) [{proc_args}]")
                try:
                    os.setresgid(gid, gid, -1)
                    os.setresuid(uid, uid, -1)
                    os.killpg(os.getpgid(pid), signum)
                except Exception as e:
                    log.warning(
                        f"User endpoint signal failed: {e} ({proc_ident}) [{proc_args}]"
                    )
                finally:
                    os.setresuid(proc_uid, proc_uid, -1)
                    os.setresgid(proc_gid, proc_gid, -1)

            deadline = time.time() + 10
            while self._child_args and time.time() < deadline:
                time.sleep(0.5)
                self.wait_for_children()

        self._command.join(5)
        log.info(
            "Shutdown complete."
            f"\n---------- Endpoint Manager ends: {self._endpoint_uuid_str}\n\n"
        )

    def _event_loop(self):
        self._command.start()

        local_user_lookup = {}
        try:
            with open("local_user_lookup.json") as f:
                local_user_lookup = json.load(f)
        except Exception as e:
            msg = (
                f"Unable to load local users ({e.__class__.__name__}) {e}\n"
                "  Will be unable to respond to any commands; update the lookup file"
                f" and either restart (stop, start) or SIGHUP ({os.getpid()}) this"
                " endpoint."
            )
            log.error(msg)

        valid_method_name_re = re.compile(r"^cmd_[A-Za-z][0-9A-Za-z_]{0,99}$")
        max_skew_s = 180  # 3 minutes; ignore commands with out-of-date timestamp
        while not self._time_to_stop:
            if self._wait_for_child:
                self.wait_for_children()

            try:
                _command = self._command_queue.get(timeout=1.0)
                d_tag, props, body = _command
                if props.headers and props.headers.get("debug", False):
                    body_log_b = _redact_url_creds(body, redact_user=False)
                    log.warning(
                        "Command debug requested:"
                        f"\n  Delivery Tag: {d_tag}"
                        f"\n  Properties: {props}"
                        f"\n  Body bytes: {body_log_b}"
                    )
            except queue.Empty:
                if self._command_stop_event.is_set():
                    self._time_to_stop = True
                if sys.stderr.isatty():
                    print(
                        f"\r{time.strftime('%c')}", end="", flush=True, file=sys.stderr
                    )
                continue

            try:
                server_cmd_ts = props.timestamp
                if props.content_type != "application/json":
                    raise ValueError("Invalid message type; expecting JSON")

                msg = json.loads(body)
                command = msg.get("command")
                command_args = msg.get("args", [])
                command_kwargs = msg.get("kwargs", {})
            except Exception as e:
                log.error(
                    f"Unable to deserialize Globus Compute services command."
                    f"  ({e.__class__.__name__}) {e}"
                )
                self._command.ack(d_tag)
                continue

            now = round(time.time())
            if abs(now - server_cmd_ts) > max_skew_s:
                server_pp_ts = datetime.fromtimestamp(server_cmd_ts).strftime("%c")
                endp_pp_ts = datetime.fromtimestamp(now).strftime("%c")
                log.warning(
                    "Ignoring command from server"
                    "\nCommand too old or skew between system clocks is too large."
                    f"\n  Command timestamp:  {server_cmd_ts:,} ({server_pp_ts})"
                    f"\n  Endpoint timestamp: {now:,} ({endp_pp_ts})"
                )
                self._command.ack(d_tag)
                continue

            try:
                globus_uuid = msg["globus_uuid"]
                globus_username = msg["globus_username"]
            except Exception as e:
                log.error(f"Invalid server command.  ({e.__class__.__name__}) {e}")
                self._command.ack(d_tag)
                continue

            try:
                local_user = local_user_lookup[globus_username]
            except Exception as e:
                log.warning(f"Invalid or unknown user.  ({e.__class__.__name__}) {e}")
                self._command.ack(d_tag)
                continue

            try:
                if not (command and valid_method_name_re.match(command)):
                    raise InvalidCommandError(f"Unknown or invalid command: {command}")

                command_func = getattr(self, command, None)
                if not command_func:
                    raise InvalidCommandError(f"Unknown or invalid command: {command}")

                command_func(local_user, command_args, command_kwargs)
                log.info(
                    f"Command process successfully forked for '{globus_username}'"
                    f" ('{globus_uuid}')."
                )
            except (InvalidCommandError, InvalidUserError) as err:
                log.error(str(err))

            except Exception:
                log.exception(
                    f"Unable to execute command: {command}\n"
                    f"    args: {command_args}\n"
                    f"  kwargs: {command_kwargs}"
                )
            finally:
                self._command.ack(d_tag)

    def cmd_start_endpoint(
        self,
        local_username: str,
        args: list[str] | None,
        kwargs: dict | None,
    ):
        if not args:
            args = []
        if not kwargs:
            kwargs = {}

        ep_name = kwargs.get("name", "")
        if not ep_name:
            raise InvalidCommandError("Missing endpoint name")

        pw_rec = pwd.getpwnam(local_username)
        udir, uid, gid = pw_rec.pw_dir, pw_rec.pw_uid, pw_rec.pw_gid
        uname = pw_rec.pw_name

        if not self._allow_same_user:
            p_uname = self._mt_user.pw_name
            if uname == p_uname or uid == os.getuid():
                raise InvalidUserError(
                    "Requested UID is same as multi-tenant UID, but configuration"
                    " has not been marked to allow the multi-tenant user to process"
                    " tasks.  To allow the same UID to also run user endpoints,"
                    " consider using a non-root user or removing privileges from UID."
                    f"\n  MT Process UID: {self._mt_user.pw_uid} ({p_uname})"
                    f"\n  Requested UID:  {uid} ({uname})",
                )

        proc_args = [
            "globus-compute-endpoint",
            "start",
            ep_name,
            "--die-with-parent",
            *args,
        ]

        try:
            pid = os.fork()
        except Exception as e:
            log.error(f"Unable to fork child process: ({e.__class__.__name__}) {e}")
            raise

        if pid > 0:
            proc_args_s = f"({uname}, {ep_name}) {' '.join(proc_args)}"
            self._child_args[pid] = (uid, gid, local_username, proc_args_s)
            log.info(f"Creating new user endpoint (pid: {pid}) [{proc_args_s}]")
            return

        # Reminder: from this point on, we are now the *child* process.
        pid = os.getpid()

        import shutil  # in the child process; no need to load this in MTEP space

        exit_code = 70
        try:
            pybindir = pathlib.Path(sys.executable).parent
            default_path = ("/usr/local/bin", "/usr/bin", "/bin", pybindir)
            env: dict[str, str] = {"PATH": ":".join(map(str, default_path))}
            env_path = self.conf_dir / "user_environment.yaml"
            try:
                if env_path.exists():
                    log.debug("Load default environment variables from: %s", env_path)
                    env_text = env_path.read_text()
                    if env_text:
                        env.update(
                            {k: str(v) for k, v in yaml.safe_load(env_text).items()}
                        )

            except Exception as e:
                log.warning(
                    "Failed to parse user environment variables from %s.  Using "
                    "default: %s\n  --- Exception ---\n(%s) %s",
                    env_path,
                    env,
                    type(e).__name__,
                    e,
                )
            env.update({"HOME": udir, "USER": uname})
            if not os.path.isdir(udir):
                udir = "/"

            wd = env.get("PWD", udir)

            os.chdir("/")  # always succeeds, so start from known place
            exit_code += 1

            if (os.getuid(), os.getgid()) != (uid, gid):
                # For multi-user systems, this is the expected path.  But for those
                # who run the multi-tenant setup as a non-privileged user, there is
                # no need to change the user: they're already executing _as that
                # uid_!
                try:
                    # The initialization of groups is "fungible" if not a
                    # privileged user
                    log.debug("Initializing groups for %s, %s", uname, gid)
                    os.initgroups(uname, gid)
                except PermissionError as e:
                    log.warning(
                        "Unable to initialize groups; unprivileged user?  Ignoring"
                        " error, but further attempts to drop privileges may fail."
                        "\n  Process ID (pid): %s"
                        "\n  Current user: %s (uid: %s, gid: %s)"
                        "\n  Attempted to initgroups to: %s (uid: %s, name: %s)",
                        os.getpid(),
                        self._mt_user.pw_name,
                        os.getuid(),
                        os.getgid(),
                        gid,
                        uid,
                        uname,
                    )
                    log.debug("Exception text: (%s) %s", e.__class__.__name__, e)
                exit_code += 1

                # But actually becoming the correct UID is _not_ fungible.  If we
                # can't -- for whatever reason -- that's a problem.  So do NOT
                # ignore the potential error.
                log.debug("Setting process group for %s to %s", pid, gid)
                os.setresgid(gid, gid, gid)  # raises (good!) on error
                exit_code += 1
                log.debug("Setting process uid for %s to %s (%s)", pid, uid, uname)
                os.setresuid(uid, uid, uid)  # raises (good!) on error
                exit_code += 1

            # If we had any capabilities, we drop them now.
            pyprctl.CapState().set_current()

            # Even if exec'ing might return some privileges, "no."
            pyprctl.set_no_new_privs()

            # some Q&D verification for admin debugging purposes
            if not shutil.which(proc_args[0], path=env["PATH"]):
                log.warning(
                    "Unable to find executable."
                    f"\n  Executable (not found): {proc_args[0]}"
                    f'\n  Path: "{env["PATH"]}"'
                    f"\n\n  Will attempt exec anyway -- WARNING - it will likely fail."
                    f"\n  (pid: {pid}, user: {uname}, {ep_name})"
                )

            os.setsid()
            exit_code += 1

            umask = 0o077  # Let child process set less restrictive, if desired
            log.debug("Setting process umask for %s to 0o%04o (%s)", pid, umask, uname)
            os.umask(umask)
            exit_code += 1

            log.debug("Changing directory to '%s'", wd)
            os.chdir(wd)
            exit_code += 1
            env["PWD"] = wd
            env["CWD"] = wd

            # in case "something gets stuck," let cmdline show it
            args_title = " ".join(proc_args)
            startup_proc_title = f"Endpoint starting up for {uname} [{args_title}]"
            setproctitle.setproctitle(startup_proc_title)

            gc_dir: pathlib.Path = GC.sdk.login_manager.tokenstore.ensure_compute_dir()
            (gc_dir / ep_name).mkdir(mode=0o700, parents=True, exist_ok=True)

            user_opts = kwargs.get("user_opts", {})
            stdin_data_dict = {
                "amqp_creds": kwargs.get("amqp_creds"),
                "config": render_config_user_template(self.conf_dir, user_opts),
            }
            stdin_data = json.dumps(stdin_data_dict)
            exit_code += 1

            # Reminder: this is *os*.open, not *open*.  Descriptors will not be closed
            # unless we explicitly do so, so `null_fd =` in loop will work.
            null_fd = os.open(os.devnull, os.O_WRONLY, mode=0o200)
            while null_fd < 3:  # reminder 0/1/2 == std in/out/err, so ...
                # ... overkill, but "just in case": don't step on them
                null_fd = os.open(os.devnull, os.O_WRONLY, mode=0o200)
            exit_code += 1

            log.debug("Setting up process stdin")
            read_handle, write_handle = os.pipe()

            # fcntl.F_GETPIPE_SZ is not available in Python versions less than 3.10
            F_GETPIPE_SZ = 1032
            # 256 - Allow some head room for multiple kernel-specific factors
            max_buf_size = fcntl.fcntl(write_handle, F_GETPIPE_SZ) - 256
            stdin_data_size = len(stdin_data)
            if stdin_data_size > max_buf_size:
                raise ValueError(
                    f"Unable to write {stdin_data_size} bytes of data to stdin; "
                    f"the maximum allowed is {max_buf_size} bytes"
                )

            exit_code += 1
            if os.dup2(read_handle, 0) != 0:  # close old stdin, use read_handle
                raise OSError("Unable to close stdin")
            os.close(read_handle)
            exit_code += 1

            log.debug("Redirecting stdout and stderr (%s)", os.devnull)
            with os.fdopen(null_fd, "w") as null_f:
                if os.dup2(null_f.fileno(), 1) != 1:
                    raise OSError("Unable to close stdout")
                exit_code += 1
                if os.dup2(null_f.fileno(), 2) != 2:
                    raise OSError("Unable to close stderr")

            # After the last os.dup2(), we are unable to get logs at *all*; hence the
            # exit_code as a last-ditch attempt at sharing "what went wrong where" to
            # the parent process.
            exit_code += 1
            log.debug("Writing credentials and config to stdin")
            with os.fdopen(write_handle, "w") as stdin_pipe:
                # intentional side effect: close handle
                stdin_pipe.write(stdin_data)

            exit_code += 1
            _soft_no, hard_no = resource.getrlimit(resource.RLIMIT_NOFILE)

            # Save closerange until last so that we can still get logs written
            # to the endpoint.log.  Meanwhile, use the exit_code as a
            # last-ditch attempt at sharing "what went wrong where" to the
            # parent process.
            exit_code += 1
            os.closerange(3, hard_no)

            exit_code += 1
            os.execvpe(proc_args[0], args=proc_args, env=env)

            # not executed, except perhaps in testing
            exit_code += 1  # type: ignore
        except Exception as e:
            log.error(f"Unable to exec for {uname} - ({e.__class__.__name__}) {e}")
        finally:
            # Only executed if execvpe fails (or isn't reached)
            exit(exit_code)



from globus_compute_endpoint.strategies.base import BaseStrategy
from globus_compute_endpoint.strategies.kube_simple import KubeSimpleStrategy
from globus_compute_endpoint.strategies.simple import SimpleStrategy

__all__ = ["BaseStrategy", "SimpleStrategy", "KubeSimpleStrategy"]



from __future__ import annotations

import argparse
import collections
import copy
import json
import logging
import os
import platform
import queue
import signal
import sys
import threading
import time
import typing as t
from collections import defaultdict

import daemon
import dill
import zmq
from globus_compute_common.messagepack.message_types import TaskTransition
from globus_compute_common.tasks import ActorName, TaskState
from globus_compute_endpoint.engines.high_throughput.interchange_task_dispatch import (  # noqa: E501
    naive_interchange_task_dispatch,
)
from globus_compute_endpoint.engines.high_throughput.messages import (
    BadCommand,
    EPStatusReport,
    Heartbeat,
    Message,
    MessageType,
)
from globus_compute_endpoint.exception_handling import (
    get_error_string,
    get_result_error_details,
)
from globus_compute_endpoint.logging_config import ComputeLogger
from globus_compute_sdk.sdk.utils import chunk_by
from globus_compute_sdk.serialize import ComputeSerializer
from parsl.version import VERSION as PARSL_VERSION

if t.TYPE_CHECKING:
    import multiprocessing as mp

log: ComputeLogger = logging.getLogger(__name__)  # type: ignore

HEARTBEAT_CODE = (2**32) - 1
PKL_HEARTBEAT_CODE = dill.dumps(HEARTBEAT_CODE)


class ManagerLost(Exception):
    """Task lost due to worker loss. Worker is considered lost when multiple heartbeats
    have been missed.
    """

    def __init__(self, worker_id):
        self.worker_id = worker_id
        self.tstamp = time.time()

    def __repr__(self):
        return f"Task failure due to loss of manager {self.worker_id}"

    def __str__(self):
        return self.__repr__()


class BadRegistration(Exception):
    """A new Manager tried to join the Engine with a BadRegistration message"""

    def __init__(self, worker_id, critical=False):
        self.worker_id = worker_id
        self.tstamp = time.time()
        self.handled = "critical" if critical else "suppressed"

    def __repr__(self):
        return (
            f"Manager {self.worker_id} attempted to register with a bad "
            f"registration message. Caused a {self.handled} failure"
        )

    def __str__(self):
        return self.__repr__()


class Interchange:
    """Interchange is a task orchestrator for distributed systems.

    1. Asynchronously queue large volume of tasks (>100K)
    2. Allow for workers to join and leave the union
    3. Detect workers that have failed using heartbeats
    4. Service single and batch requests from workers
    5. Be aware of requests worker resource capacity,
       eg. schedule only jobs that fit into walltime.

    TODO: We most likely need a PUB channel to send out global commands, like shutdown
    """

    def __init__(
        self,
        strategy=None,
        poll_period=None,
        heartbeat_period=None,
        heartbeat_threshold=1,
        working_dir=None,
        provider=None,
        max_workers_per_node=None,
        mem_per_worker=None,
        available_accelerators: t.Sequence[str] = (),
        prefetch_capacity=None,
        scheduler_mode=None,
        container_type=None,
        container_cmd_options="",
        worker_mode=None,
        cold_routing_interval=10.0,
        scaling_enabled=True,
        client_address="127.0.0.1",
        interchange_address="127.0.0.1",
        client_ports: tuple[int, int, int] = (50055, 50056, 50057),
        worker_ports=None,
        worker_port_range=None,
        cores_per_worker=1.0,
        worker_debug=False,
        launch_cmd=None,
        logdir=".",
        endpoint_id=None,
        suppress_failure=False,
    ):
        """
        Parameters
        ----------
        config : globus_compute_sdk.Config object
             Globus Compute config object that describes how compute should
             be provisioned

        client_address : str
             The ip address at which the parsl client can be reached.
             Default: "127.0.0.1"

        interchange_address : str
             The ip address at which the workers will be able to reach the Interchange.
             Default: "127.0.0.1"

        client_ports : tuple[int, int, int]
             The ports at which the client can be reached

        launch_cmd : str
             TODO : update

        worker_ports : tuple(int, int)
             The specific two ports at which workers will connect to the Interchange.
             Default: None

        worker_port_range : tuple(int, int)
             The interchange picks ports at random from the range which will be used by
             workers. This is overridden when the worker_ports option is set.
             Default: (54000, 55000)

        cores_per_worker : float
             cores to be assigned to each worker. Oversubscription is possible
             by setting cores_per_worker < 1.0. Default=1

        available_accelerators: sequence of str
            List of device IDs for accelerators available on each node
            Default: Empty list

        container_cmd_options: str
            Container command strings to be added to associated container command.
            For example, singularity exec {container_cmd_options}

        cold_routing_interval: float
            The time interval between warm and cold function routing in SOFT
            scheduler_mode.
            It is ONLY used when using soft scheduler_mode.
            We need this to avoid container workers being idle for too long.
            But we dont't want this cold routing to occur too often,
            since this may cause many warm container workers to switch to a new one.
            Default: 10.0 seconds

        worker_debug : Bool
             Enables worker debug logging.

        logdir : str
             Parsl log directory paths. Logs and temp files go here. Default: '.'

        endpoint_id : str
             Identity string that identifies the endpoint to the broker

        suppress_failure : Bool
             When set to True, the interchange will attempt to suppress failures.
             Default: False
        """

        self.logdir = logdir
        os.makedirs(self.logdir, exist_ok=True)
        log.info(f"Initializing Interchange process with Endpoint ID: {endpoint_id}")

        #
        self.max_workers_per_node = max_workers_per_node
        self.mem_per_worker = mem_per_worker
        self.cores_per_worker = cores_per_worker
        self.available_accelerators = available_accelerators
        self.prefetch_capacity = prefetch_capacity

        self.scheduler_mode = scheduler_mode
        self.container_type = container_type
        self.container_cmd_options = container_cmd_options
        self.worker_mode = worker_mode
        self.cold_routing_interval = cold_routing_interval

        self.working_dir = working_dir
        self.provider = provider
        self.worker_debug = worker_debug
        self.scaling_enabled = scaling_enabled

        self.strategy = strategy
        self.client_address = client_address
        self.interchange_address = interchange_address
        self.suppress_failure = suppress_failure

        self.poll_period = poll_period
        self.heartbeat_period = heartbeat_period
        self.heartbeat_threshold = heartbeat_threshold
        # initialize the last heartbeat time to start the loop
        self.last_heartbeat = time.time()

        self.serializer = ComputeSerializer()
        log.info(
            "Attempting connection to forwarder at {} on ports: {},{},{}".format(
                client_address, client_ports[0], client_ports[1], client_ports[2]
            )
        )
        self.context = zmq.Context()
        self.task_incoming = self.context.socket(zmq.DEALER)
        self.task_incoming.set_hwm(0)
        self.task_incoming.RCVTIMEO = 10  # in milliseconds
        log.info(f"Task incoming on tcp://{client_address}:{client_ports[0]}")
        self.task_incoming.connect(f"tcp://{client_address}:{client_ports[0]}")

        self.results_outgoing = self.context.socket(zmq.DEALER)
        self.results_outgoing.set_hwm(0)
        log.info(f"Results outgoing on tcp://{client_address}:{client_ports[1]}")
        self.results_outgoing.connect(f"tcp://{client_address}:{client_ports[1]}")

        self.command_channel = self.context.socket(zmq.DEALER)
        self.command_channel.RCVTIMEO = 1000  # in milliseconds
        # self.command_channel.set_hwm(0)
        log.info(f"Command _channel on tcp://{client_address}:{client_ports[2]}")
        self.command_channel.connect(f"tcp://{client_address}:{client_ports[2]}")
        log.info("Connected to forwarder")

        self.pending_task_queue: dict[str, queue.Queue] = {}
        self.containers: dict[str, str] = {}
        self.total_pending_task_count = 0

        log.info(f"Interchange address is {self.interchange_address}")
        self.worker_ports = worker_ports
        self.worker_port_range = (
            worker_port_range if worker_port_range is not None else (54000, 55000)
        )

        self.task_outgoing = self.context.socket(zmq.ROUTER)
        self.task_outgoing.set_hwm(0)
        self.results_incoming = self.context.socket(zmq.ROUTER)
        self.results_incoming.set_hwm(0)

        self.endpoint_id = endpoint_id
        worker_bind_address = f"tcp://{self.interchange_address}"
        log.info(f"Interchange binding worker ports to {worker_bind_address}")
        if self.worker_ports:
            self.worker_task_port = self.worker_ports[0]
            self.worker_result_port = self.worker_ports[1]

            self.task_outgoing.bind(f"{worker_bind_address}:{self.worker_task_port}")
            self.results_incoming.bind(
                f"{worker_bind_address}:{self.worker_result_port}"
            )

        else:
            self.worker_task_port = self.task_outgoing.bind_to_random_port(
                worker_bind_address,
                min_port=worker_port_range[0],
                max_port=worker_port_range[1],
                max_tries=100,
            )
            self.worker_result_port = self.results_incoming.bind_to_random_port(
                worker_bind_address,
                min_port=worker_port_range[0],
                max_port=worker_port_range[1],
                max_tries=100,
            )

        log.info(
            "Bound to ports {},{} for incoming worker connections".format(
                self.worker_task_port, self.worker_result_port
            )
        )

        self._ready_manager_queue: dict[bytes, t.Any] = {}

        self.blocks: dict[str, str] = {}
        self.block_id_map: dict[str, str] = {}
        self.launch_cmd = launch_cmd
        self.last_core_hr_counter = 0
        if not launch_cmd:
            self.launch_cmd = (
                "globus-compute-manager {debug} {max_workers} "
                "-c {cores_per_worker} "
                "--poll {poll_period} "
                "--task_url={task_url} "
                "--result_url={result_url} "
                "--logdir={logdir} "
                "--block_id={{block_id}} "
                "--hb_period={heartbeat_period} "
                "--hb_threshold={heartbeat_threshold} "
                "--worker_mode={worker_mode} "
                "--container_cmd_options='{container_cmd_options}' "
                "--scheduler_mode={scheduler_mode} "
                "--worker_type={{worker_type}} "
                "--available-accelerators {accelerator_list}"
            )

        self.current_platform = {
            "parsl_v": PARSL_VERSION,
            "python_v": "{}.{}.{}".format(
                sys.version_info.major, sys.version_info.minor, sys.version_info.micro
            ),
            "os": platform.system(),
            "hname": platform.node(),
            "dir": os.getcwd(),
        }

        log.info(f"Platform info: {self.current_platform}")
        self._block_counter = 0
        try:
            self.load_config()
        except Exception:
            log.exception("Caught exception")
            raise

        self.task_cancel_running_queue: queue.Queue = queue.Queue()
        self.task_cancel_pending_trap: dict[str, str] = {}
        self.task_status_deltas: dict[str, list[TaskTransition]] = defaultdict(list)
        self._task_status_delta_lock = threading.Lock()
        self.container_switch_count: dict[bytes, int] = {}

    def load_config(self):
        """Load the config"""
        log.info("Loading endpoint local config")
        working_dir = self.working_dir
        if self.working_dir is None:
            working_dir = os.path.join(self.logdir, "worker_logs")
        log.info(f"Setting working_dir: {working_dir}")

        self.provider.script_dir = working_dir
        if hasattr(self.provider, "channel"):
            self.provider.channel.script_dir = os.path.join(
                working_dir, "submit_scripts"
            )
            self.provider.channel.makedirs(
                self.provider.channel.script_dir, exist_ok=True
            )
            os.makedirs(self.provider.script_dir, exist_ok=True)

        debug_opts = "--debug" if self.worker_debug else ""
        max_workers = (
            ""
            if self.max_workers_per_node == float("inf")
            else f"--max_workers={self.max_workers_per_node}"
        )

        worker_task_url = f"tcp://{self.interchange_address}:{self.worker_task_port}"
        worker_result_url = (
            f"tcp://{self.interchange_address}:{self.worker_result_port}"
        )

        l_cmd = self.launch_cmd.format(
            debug=debug_opts,
            max_workers=max_workers,
            cores_per_worker=self.cores_per_worker,
            accelerator_list=" ".join(self.available_accelerators),
            # mem_per_worker=self.mem_per_worker,
            prefetch_capacity=self.prefetch_capacity,
            task_url=worker_task_url,
            result_url=worker_result_url,
            nodes_per_block=self.provider.nodes_per_block,
            heartbeat_period=self.heartbeat_period,
            heartbeat_threshold=self.heartbeat_threshold,
            poll_period=self.poll_period,
            worker_mode=self.worker_mode,
            container_cmd_options=self.container_cmd_options,
            scheduler_mode=self.scheduler_mode,
            logdir=working_dir,
        )

        self.launch_cmd = l_cmd
        log.info(f"Launch command: {self.launch_cmd}")

        if self.scaling_enabled:
            log.info("Scaling ...")
            self.scale_out(self.provider.init_blocks)

    def migrate_tasks_to_internal(self, kill_event):
        """Pull tasks from the incoming tasks 0mq pipe onto the internal
        pending task queue

        Parameters:
        -----------
        kill_event : threading.Event
              Event to let the thread know when it is time to die.
        """
        log.info("Starting")
        task_counter = 0
        poller = zmq.Poller()
        poller.register(self.task_incoming, zmq.POLLIN)

        while not kill_event.is_set():
            # We are no longer doing heartbeats on the task side.
            try:
                raw_msg = self.task_incoming.recv()
                self.last_heartbeat = time.time()
            except zmq.Again:
                log.trace(
                    "No new incoming task - %s tasks in internal queue",
                    self.total_pending_task_count,
                )
                continue

            try:
                msg = Message.unpack(raw_msg)
            except Exception:
                log.exception(f"Failed to unpack message, RAW:{raw_msg}")
                continue

            if msg == "STOP":
                # TODO: Yadu. This should be replaced by a proper MessageType
                log.debug("Received STOP message.")
                kill_event.set()
                break
            elif isinstance(msg, Heartbeat):
                log.debug("Got heartbeat")
            else:
                log.info(f"Received task: {msg.task_id}")
                local_container = msg.container_id
                self.containers[local_container] = local_container
                msg.set_local_container(local_container)
                if local_container not in self.pending_task_queue:
                    self.pending_task_queue[local_container] = queue.Queue(
                        maxsize=10**6
                    )

                # We pass the raw message along
                self.pending_task_queue[local_container].put(
                    {
                        "task_id": msg.task_id,
                        "container_id": msg.container_id,
                        "local_container": local_container,
                        "raw_buffer": raw_msg,
                    }
                )
                self.total_pending_task_count += 1
                tt = TaskTransition(
                    timestamp=time.time_ns(),
                    state=TaskState.WAITING_FOR_NODES,
                    actor=ActorName.INTERCHANGE,
                )

                with self._task_status_delta_lock:
                    self.task_status_deltas[msg.task_id].append(tt)

                log.debug(
                    f"[TASK_PULL_THREAD] task {msg.task_id} is now WAITING_FOR_NODES"
                )
                log.debug(
                    "[TASK_PULL_THREAD] pending task count: {}".format(
                        self.total_pending_task_count
                    )
                )
                task_counter += 1
                log.debug(f"[TASK_PULL_THREAD] Fetched task:{task_counter}")

    def get_total_tasks_outstanding(self):
        """Get the outstanding tasks in total"""
        outstanding = {}
        for task_type in self.pending_task_queue:
            outstanding[task_type] = (
                outstanding.get(task_type, 0)
                + self.pending_task_queue[task_type].qsize()
            )
        for manager in self._ready_manager_queue:
            for task_type in self._ready_manager_queue[manager]["tasks"]:
                outstanding[task_type] = outstanding.get(task_type, 0) + len(
                    self._ready_manager_queue[manager]["tasks"][task_type]
                )
        return outstanding

    def get_total_live_workers(self):
        """Get the total active workers"""
        active = 0
        for manager in self._ready_manager_queue:
            if self._ready_manager_queue[manager]["active"]:
                active += self._ready_manager_queue[manager]["max_worker_count"]
        return active

    def get_outstanding_breakdown(self):
        """Get outstanding breakdown per manager and in the interchange queues

        Returns
        -------
        List of status for online elements
        [ (element, tasks_pending, status) ... ]
        """

        pending_on_interchange = self.total_pending_task_count
        # Reporting pending on interchange is a deviation from Parsl
        reply = [("interchange", pending_on_interchange, True)]
        for manager in self._ready_manager_queue:
            resp = (
                manager.decode("utf-8"),
                sum(
                    len(tids)
                    for tids in self._ready_manager_queue[manager]["tasks"].values()
                ),
                self._ready_manager_queue[manager]["active"],
            )
            reply.append(resp)
        return reply

    def _hold_block(self, block_id):
        """Sends hold command to all managers which are in a specific block

        Parameters
        ----------
        block_id : str
             Block identifier of the block to be put on hold
        """
        for manager in self._ready_manager_queue:
            if (
                self._ready_manager_queue[manager]["active"]
                and self._ready_manager_queue[manager]["block_id"] == block_id
            ):
                log.debug(f"[HOLD_BLOCK]: Sending hold to manager: {manager}")
                self.hold_manager(manager)

    def hold_manager(self, manager):
        """Put manager on hold
        Parameters
        ----------

        manager : str
          Manager id to be put on hold while being killed
        """
        if manager in self._ready_manager_queue:
            self._ready_manager_queue[manager]["active"] = False

    def _status_report_loop(self, kill_event, status_report_queue: queue.Queue):
        log.info(f"Endpoint id: {self.endpoint_id}")

        def _enqueue_status_report(ep_state: dict, task_states: dict):
            try:
                msg = EPStatusReport(str(self.endpoint_id), ep_state, dict(task_states))
                status_report_queue.put(msg.pack())
            except Exception:
                log.exception("Unable to create or send EP status report.")
                log.debug("Attempted to send chunk: %s", tsd_chunk)
                # ignoring so that the thread continues; "it's just a status"

        while True:
            with self._task_status_delta_lock:
                task_status_deltas = copy.deepcopy(self.task_status_deltas)
                self.task_status_deltas.clear()

            log.debug(
                "Cleared task deltas (%s); sending status report to Engine.",
                len(task_status_deltas),
            )

            # For multi-chunk reports ("lots-o-tasks"), the state won't change *that*
            # much each iteration, so cache the result
            global_state = self.get_global_state_for_status_report()

            # The result processor will gracefully handle any size message, but
            # courtesy says to chunk work; 4,096 is empirically chosen to be plenty
            # "bulk enough," but not rude.
            for tsd_chunk in chunk_by(task_status_deltas.items(), 4_096):
                _enqueue_status_report(global_state, dict(tsd_chunk))

            if not task_status_deltas:
                _enqueue_status_report(global_state, {})

            del task_status_deltas  # free some memory in "large case"

            if kill_event.wait(self.heartbeat_period):
                break

    def _command_server(self, kill_event):
        """Command server to run async command to the interchange

        We want to be able to receive the following not yet implemented/updated
        commands:
         - OutstandingCount
         - ListManagers (get outstanding broken down by manager)
         - HoldWorker
         - Shutdown
        """
        log.debug("Command Server Starting")

        while not kill_event.is_set():
            try:
                buffer = self.command_channel.recv()
                log.debug(f"Received command request {buffer}")
                command = Message.unpack(buffer)

                if command.type is MessageType.TASK_CANCEL:
                    log.info(f"Received TASK_CANCEL for Task:{command.task_id}")
                    self.enqueue_task_cancel(command.task_id)
                    reply = command

                elif command.type is MessageType.HEARTBEAT_REQ:
                    log.info("Received synchonous HEARTBEAT_REQ from hub")
                    log.info(f"Replying with Heartbeat({self.endpoint_id})")
                    reply = Heartbeat(self.endpoint_id)

                else:
                    log.error(
                        f"Received unsupported message type:{command.type} on "
                        "command _channel"
                    )
                    reply = BadCommand(f"Unknown command type: {command.type}")

                log.debug(f"Reply: {reply}")
                self.command_channel.send(reply.pack())

            except zmq.Again:
                log.trace("Command server is alive")
                continue

    def enqueue_task_cancel(self, task_id):
        """Cancel a task on the interchange
        Here are the task states and responses we issue here
        1. Task is pending in queues -> we add task to a trap to capture while in
            dispatch and delegate cancel to the manager the task is assigned to
        2. Task is in a transitionary state between pending in queue and dispatched ->
               task is added pre-emptively to trap
        3. Task is pending on a manager -> we delegate cancellation to manager
        4. Task is already complete -> we leave in trap, since we can't know

        We place the task in the trap so that even if the search misses, the task
        will be caught from getting dispatched even if the search fails due to a
        race-condition. Since the task can't be dispatched before scheduling is
        complete, either must work.
        """
        log.debug(f"Received task_cancel request for Task:{task_id}")

        self.task_cancel_pending_trap[task_id] = task_id
        for manager in self._ready_manager_queue:
            for task_type in self._ready_manager_queue[manager]["tasks"]:
                for tid in self._ready_manager_queue[manager]["tasks"][task_type]:
                    if tid == task_id:
                        log.debug(
                            f"Task:{task_id} is running, "
                            "moving task_cancel message onto queue"
                        )
                        self.task_cancel_running_queue.put((manager, task_id))
                        self.task_cancel_pending_trap.pop(task_id, None)
                        break
        return

    def handle_sigterm(self, sig_num, curr_stack_frame):
        log.warning("Received SIGTERM, stopping")
        self.stop()

    def stop(self):
        """Prepare the interchange for shutdown"""
        self._kill_event.set()

        self._task_puller_thread.join()
        self._command_thread.join()
        self._status_report_thread.join()
        log.debug("HighThroughput Interchange stopped")

    def start(self, poll_period: int | None = None) -> None:
        """Start the Interchange

        Parameters:
        ----------
        poll_period : int
           poll_period in milliseconds
        """
        signal.signal(signal.SIGTERM, self.handle_sigterm)
        log.info("Incoming ports bound")

        if poll_period is None:
            poll_period = self.poll_period

        start = time.time()
        count = 0

        self._kill_event = threading.Event()
        self._task_puller_thread = threading.Thread(
            target=self.migrate_tasks_to_internal,
            args=(self._kill_event,),
            name="TASK_PULL_THREAD",
        )
        self._task_puller_thread.start()

        self._command_thread = threading.Thread(
            target=self._command_server, args=(self._kill_event,), name="COMMAND_THREAD"
        )
        self._command_thread.start()

        status_report_queue: queue.Queue[bytes] = queue.Queue()
        self._status_report_thread = threading.Thread(
            target=self._status_report_loop,
            args=(self._kill_event, status_report_queue),
            name="STATUS_THREAD",
        )
        self._status_report_thread.start()

        try:
            log.info("Starting strategy.")
            self.strategy.start(self)
        except RuntimeError:
            # This is raised when re-registering an endpoint as strategy already exists
            log.exception("Failed to start strategy.")

        poller = zmq.Poller()
        # poller.register(self.task_incoming, zmq.POLLIN)
        poller.register(self.task_outgoing, zmq.POLLIN)
        poller.register(self.results_incoming, zmq.POLLIN)

        # These are managers which we should examine in an iteration
        # for scheduling a job (or maybe any other attention?).
        # Anything altering the state of the manager should add it
        # onto this list.
        interesting_managers: set[bytes] = set()

        # This value records when the last cold routing in soft mode happens
        # When the cold routing in soft mode happens, it may cause worker containers to
        # switch
        # Cold routing is to reduce the number idle workers of specific task types on
        # the managers when there are not enough tasks of those types in the task queues
        # on interchange
        last_cold_routing_time = time.time()
        prev_manager_stat = None

        task_deltas_to_merge: dict[str, list[TaskTransition]] = defaultdict(list)

        while not self._kill_event.is_set():
            self.socks = dict(poller.poll(timeout=poll_period))

            # Listen for requests for work
            if (
                self.task_outgoing in self.socks
                and self.socks[self.task_outgoing] == zmq.POLLIN
            ):
                log.trace("starting task_outgoing section")
                message = self.task_outgoing.recv_multipart()
                manager = message[0]

                mdata = self._ready_manager_queue.get(manager)
                if not mdata:
                    reg_flag = False

                    try:
                        msg = json.loads(message[1].decode("utf-8"))
                        reg_flag = True
                    except Exception:
                        log.warning(
                            "Got a non-json registration message from manager:%s",
                            manager,
                        )
                        log.debug("Message :\n%s\n", message)

                    # By default we set up to ignore bad nodes/registration messages.
                    now = time.time()
                    mdata = {
                        "last": now,
                        "reg_time": now,
                        "free_capacity": {"total_workers": 0},
                        "max_worker_count": 0,
                        "active": True,
                        "tasks": collections.defaultdict(set),
                        "total_tasks": 0,
                    }
                    if reg_flag is True:
                        interesting_managers.add(manager)
                        log.info(
                            f"Add manager to ready queue: {manager!r}"
                            f"\n  Registration info: {msg})"
                        )
                        mdata.update(msg)
                        self._ready_manager_queue[manager] = mdata

                        if (
                            msg["python_v"].rsplit(".", 1)[0]
                            != self.current_platform["python_v"].rsplit(".", 1)[0]
                            or msg["parsl_v"] != self.current_platform["parsl_v"]
                        ):
                            log.info(
                                f"Manager:{manager!r} version:{msg['python_v']} "
                                "does not match the interchange"
                            )
                    else:
                        # Registration has failed.
                        if self.suppress_failure is False:
                            log.debug("Setting kill event for bad manager")
                            self._kill_event.set()
                            e = BadRegistration(manager, critical=True)
                            result_package = {
                                "task_id": -1,
                                "exception": self.serializer.serialize(e),
                            }
                            pkl_package = dill.dumps(result_package)
                            self.results_outgoing.send(dill.dumps([pkl_package]))
                        else:
                            log.debug(
                                "Suppressing bad registration from manager: %s",
                                manager,
                            )

                else:
                    mdata["last"] = time.time()
                    if message[1] == b"HEARTBEAT":
                        log.debug("Manager %s sends heartbeat", manager)
                        self.task_outgoing.send_multipart(
                            [manager, b"", PKL_HEARTBEAT_CODE]
                        )
                    else:
                        manager_adv = dill.loads(message[1])
                        log.debug("Manager %s requested %s", manager, manager_adv)
                        manager_adv["total_workers"] = sum(manager_adv["free"].values())
                        mdata["free_capacity"].update(manager_adv)
                        interesting_managers.add(manager)
                        del manager_adv

            # If we had received any requests, check if there are tasks that could be
            # passed

            cur_manager_stat = len(self._ready_manager_queue), len(interesting_managers)
            if cur_manager_stat != prev_manager_stat:
                prev_manager_stat = cur_manager_stat
                _msg = "[MAIN] New managers count (total/interesting): {}/{}"
                log.debug(_msg.format(*cur_manager_stat))

            if time.time() - last_cold_routing_time > self.cold_routing_interval:
                task_dispatch, dispatched_task = naive_interchange_task_dispatch(
                    interesting_managers,
                    self.pending_task_queue,
                    self._ready_manager_queue,
                    scheduler_mode=self.scheduler_mode,
                    cold_routing=True,
                )
                last_cold_routing_time = time.time()
            else:
                task_dispatch, dispatched_task = naive_interchange_task_dispatch(
                    interesting_managers,
                    self.pending_task_queue,
                    self._ready_manager_queue,
                    scheduler_mode=self.scheduler_mode,
                    cold_routing=False,
                )

            self.total_pending_task_count -= dispatched_task

            # Task cancel is high priority, so we'll process all requests
            # in one go
            try:
                while True:
                    manager, task_id = self.task_cancel_running_queue.get(block=False)
                    log.debug(
                        "CANCELLED running task (id: %s, manager: %s)", task_id, manager
                    )
                    cancel_message = dill.dumps(("TASK_CANCEL", task_id))
                    self.task_outgoing.send_multipart([manager, b"", cancel_message])
            except queue.Empty:
                pass

            for manager in task_dispatch:
                tasks = task_dispatch[manager]
                if tasks:
                    log.info(
                        'Sending task message "{}..." to manager {!r}'.format(
                            str(tasks)[:50], manager
                        )
                    )
                    serializd_raw_tasks_buffer = dill.dumps(tasks)
                    self.task_outgoing.send_multipart(
                        [manager, b"", serializd_raw_tasks_buffer]
                    )

                    for task in tasks:
                        task_id = task["task_id"]
                        log.info(f"Sent task {task_id} to manager {manager!r}")
                        if (
                            self.task_cancel_pending_trap
                            and task_id in self.task_cancel_pending_trap
                        ):
                            log.info(f"Task:{task_id} CANCELLED before launch")
                            cancel_message = dill.dumps(("TASK_CANCEL", task_id))
                            self.task_outgoing.send_multipart(
                                [manager, b"", cancel_message]
                            )
                            self.task_cancel_pending_trap.pop(task_id)
                        else:
                            log.debug("Task:%s is now WAITING_FOR_LAUNCH", task_id)
                            tt = TaskTransition(
                                timestamp=time.time_ns(),
                                state=TaskState.WAITING_FOR_LAUNCH,
                                actor=ActorName.INTERCHANGE,
                            )
                            task_deltas_to_merge[task_id].append(tt)

            if task_deltas_to_merge:
                with self._task_status_delta_lock:
                    for task_id, deltas in task_deltas_to_merge.items():
                        self.task_status_deltas[task_id].extend(deltas)
                task_deltas_to_merge.clear()

            # Receive any results and forward to client
            if (
                self.results_incoming in self.socks
                and self.socks[self.results_incoming] == zmq.POLLIN
            ):
                log.debug("entering results_incoming section")
                manager, *b_messages = self.results_incoming.recv_multipart()
                mdata = self._ready_manager_queue.get(manager)
                if not mdata:
                    log.warning(
                        "Received a result from a un-registered manager: %s",
                        manager,
                    )
                else:
                    # We expect the batch of messages to be (optionally) a task status
                    # update message followed by 0 or more task results
                    try:
                        log.debug("Trying to unpack")
                        manager_report = Message.unpack(b_messages[0])
                        if manager_report.task_statuses:
                            log.info(
                                "Got manager status report: %s",
                                manager_report.task_statuses,
                            )

                            for tid, statuses in manager_report.task_statuses.items():
                                task_deltas_to_merge[tid].extend(statuses)

                        self.task_outgoing.send_multipart(
                            [manager, b"", PKL_HEARTBEAT_CODE]
                        )
                        b_messages = b_messages[1:]
                        mdata["last"] = time.time()
                        self.container_switch_count[
                            manager
                        ] = manager_report.container_switch_count
                        log.info(
                            "Got container switch count: %s",
                            self.container_switch_count,
                        )
                    except Exception:
                        pass
                    if len(b_messages):
                        log.info(f"Got {len(b_messages)} result items in batch")
                        for idx, b_message in enumerate(b_messages):
                            r = dill.loads(b_message)
                            tid = r["task_id"]

                            log.debug("Received task result %s (from %s)", tid, manager)
                            task_container = self.containers[r["container_id"]]
                            log.debug(
                                "Removing for manager: %s from %s",
                                manager,
                                self._ready_manager_queue,
                            )

                            mdata["tasks"][task_container].remove(tid)
                            b_messages[idx] = dill.dumps(r)

                        mdata["total_tasks"] -= len(b_messages)

                    self.results_outgoing.send(dill.dumps(b_messages))
                    interesting_managers.add(manager)

                    log.debug(f"Current tasks: {mdata['tasks']}")
                log.debug("leaving results_incoming section")

            # Send status reports from this main thread to avoid thread-safety on zmq
            # sockets
            try:
                packed_status_report = status_report_queue.get(block=False)
                log.trace("forwarding status report: %s", packed_status_report)
                self.results_outgoing.send(packed_status_report)
            except queue.Empty:
                pass

            now = time.time()
            hbt_window_start = now - self.heartbeat_threshold
            bad_managers = [
                manager
                for manager, mdata in self._ready_manager_queue.items()
                if hbt_window_start > mdata["last"]
            ]
            bad_manager_msgs = []
            for manager in bad_managers:
                log.debug(
                    "Last: %s Current: %s",
                    self._ready_manager_queue[manager]["last"],
                    now,
                )
                log.warning(f"Too many heartbeats missed for manager {manager!r}")
                for tasks in self._ready_manager_queue[manager]["tasks"].values():
                    for tid in tasks:
                        try:
                            raise ManagerLost(manager)
                        except Exception:
                            result_package = {
                                "task_id": tid,
                                "exception": get_error_string(),
                                "error_details": get_result_error_details(),
                            }
                            pkl_package = dill.dumps(result_package)
                            bad_manager_msgs.append(pkl_package)
                log.warning(f"Unregistering manager {manager!r}")
                self._ready_manager_queue.pop(manager, None)
                if manager in interesting_managers:
                    interesting_managers.remove(manager)
            if bad_manager_msgs:
                log.warning(f"Sending task failure reports of manager {manager!r}")
                self.results_outgoing.send(dill.dumps(bad_manager_msgs))

        delta = time.time() - start
        log.info(f"Processed {count} tasks in {delta} seconds")
        log.warning("Exiting")

    def get_global_state_for_status_report(self):
        outstanding_tasks = self.get_total_tasks_outstanding()
        pending_tasks = self.total_pending_task_count
        num_managers = len(self._ready_manager_queue)
        live_workers = self.get_total_live_workers()
        free_capacity = sum(
            m["free_capacity"]["total_workers"]
            for m in self._ready_manager_queue.values()
        )

        return {
            "managers": num_managers,
            "total_workers": live_workers,
            "idle_workers": free_capacity,
            "pending_tasks": pending_tasks,
            "outstanding_tasks": outstanding_tasks,
            "heartbeat_period": self.heartbeat_period,
        }

    def scale_out(self, blocks=1, task_type=None):
        """Scales out the number of blocks by "blocks"

        Raises:
             NotImplementedError
        """
        log.info(f"Scaling out by {blocks} more blocks for task type {task_type}")
        r = []
        for _i in range(blocks):
            if self.provider:
                self._block_counter += 1
                external_block_id = str(self._block_counter)
                if not task_type and self.scheduler_mode == "hard":
                    launch_cmd = self.launch_cmd.format(
                        block_id=external_block_id, worker_type="RAW"
                    )
                else:
                    launch_cmd = self.launch_cmd.format(
                        block_id=external_block_id, worker_type=task_type
                    )
                if not task_type:
                    internal_block = self.provider.submit(launch_cmd, 1)
                else:
                    internal_block = self.provider.submit(launch_cmd, 1, task_type)
                log.debug(f"Launched block {external_block_id}->{internal_block}")
                if not internal_block:
                    raise RuntimeError(
                        "Attempt to provision nodes via provider "
                        f"{self.provider.label} has failed"
                    )
                self.blocks[external_block_id] = internal_block
                self.block_id_map[internal_block] = external_block_id
            else:
                log.error("No execution provider available")
                r = None
        return r

    def scale_in(self, blocks=None, block_ids=None, task_type=None):
        """Scale in the number of active blocks by specified amount.

        Parameters
        ----------
        blocks : int
            # of blocks to terminate

        block_ids : [str.. ]
            List of external block ids to terminate
        """
        if block_ids is None:
            block_ids = []
        if task_type:
            log.info(
                "Scaling in blocks of specific task type %s. Let the provider decide "
                "which to kill",
                task_type,
            )
            if self.scaling_enabled and self.provider:
                to_kill, r = self.provider.cancel(blocks, task_type)
                log.info(f"Get the killed blocks: {to_kill}, and status: {r}")
                for job in to_kill:
                    log.info(
                        "[scale_in] Getting the block_id map {} for job {}".format(
                            self.block_id_map, job
                        )
                    )
                    block_id = self.block_id_map[job]
                    log.info(f"[scale_in] Holding block {block_id}")
                    self._hold_block(block_id)
                    self.blocks.pop(block_id)
                return r

        if block_ids:
            block_ids_to_kill = block_ids
        else:
            block_ids_to_kill = list(self.blocks.keys())[:blocks]

        # Try a polite terminate
        # TODO : Missing logic to hold blocks
        for block_id in block_ids_to_kill:
            self._hold_block(block_id)

        # Now kill via provider
        to_kill = [self.blocks.pop(bid) for bid in block_ids_to_kill]

        if self.scaling_enabled and self.provider:
            r = self.provider.cancel(to_kill)

        return r

    def provider_status(self):
        """Get status of all blocks from the provider. The return type is
        defined by the particular provider in use.
        """
        status = []
        if self.provider:
            job_ids: list[str] = list(self.blocks.values())
            log.trace("Getting the status of %s blocks.", job_ids)
            status = self.provider.status(job_ids)
            log.trace("The status is %s", status)

        return status


def starter(comm_q: mp.Queue, *args, **kwargs) -> None:
    """Start the interchange process

    The Engine is expected to call this function. The args, kwargs match that of the
    Interchange.__init__
    """
    ic = None
    try:
        ic = Interchange(*args, **kwargs)
        comm_q.put((ic.worker_task_port, ic.worker_result_port))
    finally:
        if not ic:  # There was an exception
            comm_q.put(None)

        # no sense in having the queue open past it's usefulness
        comm_q.close()
        comm_q.join_thread()
        del comm_q

    if ic:
        ic.start()


def cli_run():
    from globus_compute_endpoint.logging_config import setup_logging

    parser = argparse.ArgumentParser()
    parser.add_argument("-c", "--client_address", required=True, help="Client address")
    parser.add_argument(
        "--client_ports",
        required=True,
        help="client ports as a triple of outgoing,incoming,command",
    )
    parser.add_argument("--worker_port_range", help="Worker port range as a tuple")
    parser.add_argument(
        "-l",
        "--logdir",
        default="./parsl_worker_logs",
        help="Parsl worker log directory",
    )
    parser.add_argument(
        "-p", "--poll_period", help="REQUIRED: poll period used for main thread"
    )
    parser.add_argument(
        "--worker_ports",
        default=None,
        help="OPTIONAL, pair of workers ports to listen on, "
        "eg --worker_ports=50001,50005",
    )
    parser.add_argument(
        "--suppress_failure",
        action="store_true",
        help="Enables suppression of failures",
    )
    parser.add_argument(
        "--endpoint_id",
        default=None,
        help="Endpoint ID, used to identify the endpoint to the remote broker",
    )
    parser.add_argument("--hb_threshold", help="Heartbeat threshold in seconds")
    parser.add_argument(
        "--config",
        default=None,
        help="Configuration object that describes provisioning",
    )
    parser.add_argument(
        "-d", "--debug", action="store_true", help="Enables debug logging"
    )

    print("Starting HTEX Intechange")

    args = parser.parse_args()

    args.logdir = os.path.abspath(args.logdir)
    if args.worker_ports:
        args.worker_ports = [int(i) for i in args.worker_ports.split(",")]
    if args.worker_port_range:
        args.worker_port_range = [int(i) for i in args.worker_port_range.split(",")]

    setup_logging(
        logfile=os.path.join(args.logdir, "interchange.log"),
        debug=args.debug,
        console_enabled=False,
    )

    with daemon.DaemonContext():
        ic = Interchange(
            logdir=args.logdir,
            suppress_failure=args.suppress_failure,
            client_address=args.client_address,
            client_ports=[int(i) for i in args.client_ports.split(",")],
            endpoint_id=args.endpoint_id,
            config=args.config,
            worker_ports=args.worker_ports,
            worker_port_range=args.worker_port_range,
        )
        ic.start()



from __future__ import annotations

import importlib.util
import inspect
import json
import logging
import pathlib
import shlex

import yaml
from click import ClickException
from packaging.version import Version
from pydantic import ValidationError

from .config import Config
from .model import ConfigModel

log = logging.getLogger(__name__)


def _load_config_py(conf_path: pathlib.Path) -> Config | None:
    if not conf_path.exists():
        return None

    try:
        from funcx_endpoint.version import VERSION

        if Version(VERSION) < Version("2.0.0"):
            msg = (
                "To avoid compatibility issues with Globus Compute, please uninstall "
                "funcx-endpoint or upgrade funcx-endpoint to >=2.0.0. Note that the "
                "funcx-endpoint package is now deprecated."
            )
            raise ClickException(msg)
    except ModuleNotFoundError:
        pass

    try:
        spec = importlib.util.spec_from_file_location("config", conf_path)
        if not (spec and spec.loader):
            raise Exception(f"Unable to import configuration (no spec): {conf_path}")
        config = importlib.util.module_from_spec(spec)
        if not config:
            raise Exception(f"Unable to import configuration (no config): {conf_path}")
        spec.loader.exec_module(config)
        return config.config

    except FileNotFoundError as err:
        msg = (
            f"{err}"
            "\n\nUnable to find required configuration file; has the configuration"
            "\ndirectory been corrupted?"
        )
        raise ClickException(msg) from err

    except AttributeError as err:
        msg = (
            f"{err}"
            "\n\nFailed to find expected data structure in configuration file."
            "\nHas the configuration file been corrupted or modified incorrectly?\n"
        )
        raise ClickException(msg) from err

    except ModuleNotFoundError as err:
        # Catch specific error when old config.py references funcx_endpoint
        if "No module named 'funcx_endpoint'" in err.msg:
            msg = (
                f"{conf_path} contains import statements from a previously "
                "configured endpoint that uses the (deprecated) "
                "funcx-endpoint library. Please update the imports to reference "
                "globus_compute_endpoint.\n\ni.e.\n"
                "    from funcx_endpoint.endpoint.utils.config -> "
                "from globus_compute_endpoint.endpoint.config\n"
                "    from funcx_endpoint.executors -> "
                "from globus_compute_endpoint.executors\n"
                "\n"
                "You can also use the command "
                "`globus-compute-endpoint update_funcx_config [endpoint_name]` "
                "to update them\n"
            )
            raise ClickException(msg) from err
        else:
            log.exception(err.msg)
            raise

    except Exception:
        log.exception(
            "Globus Compute v2.0.0 made several non-backwards compatible changes to "
            "the config. Your config might be out of date. "
            "Refer to "
            "https://funcx.readthedocs.io/en/latest/endpoints.html#configuring-funcx"
        )
        raise


def _read_config_yaml(config_path: pathlib.Path) -> str:
    return config_path.read_text()


def load_config_yaml(config_str: str) -> Config:
    try:
        config_dict = dict(yaml.safe_load(config_str))
    except Exception as err:
        raise ClickException(f"Invalid config syntax: {str(err)}") from err

    try:
        config_schema = ConfigModel(**config_dict)
    except ValidationError as err:
        raise ClickException(str(err)) from err

    config_dict = config_schema.dict(exclude_unset=True)

    try:
        config = Config(**config_dict)
    except Exception as err:
        raise ClickException(str(err)) from err

    # Special case of 'display_name: None' converting to the empty
    #   None instead of the string "None" which was the default
    #   display_name since we started generating yaml configs from
    #   Jun til Jul 2023.
    # This has the side effect of disallowing the display name 'None'
    if config.display_name == "None":
        config.display_name = None
    return config


def get_config(endpoint_dir: pathlib.Path) -> Config:
    config_py_path = endpoint_dir / "config.py"
    config_yaml_path = endpoint_dir / "config.yaml"

    config = _load_config_py(config_py_path)
    if config:
        # Use config.py if present
        # Note that if config.py exists but is invalid,
        # an exception will be raised
        return config

    try:
        config_str = _read_config_yaml(config_yaml_path)
    except FileNotFoundError as err:
        endpoint_name = endpoint_dir.name

        if endpoint_dir.exists():
            msg = (
                f"{err}"
                "\n\nUnable to find required configuration file; has the configuration"
                "\ndirectory been corrupted?"
            )
        else:
            configure_command = "globus-compute-endpoint configure"
            if endpoint_name != "default":
                configure_command += f" {endpoint_name}"
            msg = (
                f"{err}"
                f"\n\nEndpoint '{endpoint_name}' is not configured!"
                "\n1. Please create a configuration template with:"
                f"\n\t{configure_command}"
                "\n2. Update the configuration"
                "\n3. Start the endpoint\n"
            )
        raise ClickException(msg) from err

    config = load_config_yaml(config_str)
    return config


def load_user_config_schema(endpoint_dir: pathlib.Path) -> dict | None:
    from globus_compute_endpoint.endpoint.endpoint import Endpoint

    user_config_schema_path = Endpoint.user_config_schema_path(endpoint_dir)
    if not user_config_schema_path.exists():
        return None

    try:
        return json.loads(user_config_schema_path.read_text())
    except json.JSONDecodeError:
        log.error(
            f"\n\nThe user config schema is not valid JSON: {user_config_schema_path}\n"
        )
        raise


def _validate_user_opts(user_opts: dict, schema: dict) -> None:
    """Validates user config options against a JSON schema."""
    import jsonschema  # Only load package when called by EP manager

    try:
        jsonschema.validate(instance=user_opts, schema=schema)
    except jsonschema.SchemaError:
        log.error("\n\nThe user config schema is invalid\n")
        raise
    except jsonschema.ValidationError:
        log.error("\n\nThe provided user config options are invalid\n")
        raise


def _sanitize_user_opts(data):
    """Prevent YAML injection via special characters.
    Note that this will enforce double quoted YAML strings.
    """
    if isinstance(data, dict):
        return {k: _sanitize_user_opts(v) for k, v in data.items()}
    elif isinstance(data, list):
        return [_sanitize_user_opts(v) for v in data]
    elif isinstance(data, str):
        return json.dumps(data)
    elif isinstance(data, (int, float)):
        return data
    else:
        # We do not expect to hit this because user options are passed
        # from the web service as JSON
        raise ValueError(
            f"{type(data).__name__} is not a valid user config option type"
        )


def _shell_escape_filter(val):
    """Returns a shell-escaped version of the argument"""
    if not isinstance(val, str):
        return val

    # We need to escape the command before serializing into JSON
    # because PyYAML will strip the surrounding quotes when the
    # YAML content is loaded
    loaded = json.loads(val)
    if not isinstance(loaded, str):
        return val

    return json.dumps(shlex.quote(loaded))


def render_config_user_template(endpoint_dir: pathlib.Path, user_opts: dict) -> str:
    import jinja2  # Only load package when called by EP manager
    from globus_compute_endpoint.endpoint.endpoint import Endpoint
    from jinja2.sandbox import SandboxedEnvironment

    user_config_schema = load_user_config_schema(endpoint_dir)
    if user_config_schema:
        _validate_user_opts(user_opts, user_config_schema)
    user_opts = _sanitize_user_opts(user_opts)
    user_config_path = Endpoint.user_config_template_path(endpoint_dir)

    template_str = _read_config_yaml(user_config_path)
    environment = SandboxedEnvironment(undefined=jinja2.StrictUndefined)
    environment.filters["shell_escape"] = _shell_escape_filter
    template = environment.from_string(template_str)

    try:
        return template.render(**user_opts)
    except jinja2.exceptions.UndefinedError as e:
        log.debug(f"Missing required user option(s): {e}")
        raise
    except jinja2.exceptions.SecurityError as e:
        log.debug(f"Template tried accessing insecure code: {e}")
        raise


def serialize_config(config: Config) -> dict:
    """Converts a Config object into a dict that matches the
    standard YAML config schema.

    E.g.,
    {
        ...
        "display_name": "My Endpoint",
        ...
        "executor": {
            ...
            "provider": {
                ...
                "type": "LocalProvider"
                ...
            }
            ...
        }
    }
    """

    def _prep(val):
        if hasattr(val, "__dict__") and hasattr(val, "__init__"):
            return _to_dict(val)
        try:
            json.dumps(val, allow_nan=False)
            return val
        except (TypeError, ValueError):
            return repr(val)

    def _to_dict(obj):
        res = {"type": type(obj).__name__}

        # We only want to include attributes that are
        # configurable as constructor arguments
        sig = inspect.signature(obj.__init__)

        for param in sig.parameters.keys():
            val = getattr(obj, param, 0)
            if val == 0:
                continue

            if isinstance(val, (list, tuple)):
                res[param] = [_prep(item) for item in val]

            else:
                res[param] = _prep(val)

        return res

    return _to_dict(config)



from __future__ import annotations

import json
import uuid
from abc import ABC, abstractmethod
from collections import defaultdict
from enum import Enum, auto
from struct import Struct

from globus_compute_common.messagepack.message_types import TaskTransition

MESSAGE_TYPE_FORMATTER = Struct("b")


class MessageType(Enum):
    HEARTBEAT_REQ = auto()
    HEARTBEAT = auto()
    EP_STATUS_REPORT = auto()
    MANAGER_STATUS_REPORT = auto()
    TASK = auto()
    RESULTS_ACK = auto()
    TASK_CANCEL = auto()
    BAD_COMMAND = auto()

    def pack(self):
        return MESSAGE_TYPE_FORMATTER.pack(self.value)

    @classmethod
    def unpack(cls, buffer):
        (mtype,) = MESSAGE_TYPE_FORMATTER.unpack_from(buffer, offset=0)
        return MessageType(mtype), buffer[MESSAGE_TYPE_FORMATTER.size :]


COMMAND_TYPES = {MessageType.HEARTBEAT_REQ, MessageType.TASK_CANCEL}


class Message(ABC):
    def __init__(self):
        self._payload = None
        self._header = None

    @property
    def header(self):
        return self._header

    @property
    def type(self):
        raise NotImplementedError()

    @property
    def payload(self):
        return self._payload

    @classmethod
    def unpack(cls, msg):
        message_type, remaining = MessageType.unpack(msg)
        if message_type is MessageType.HEARTBEAT_REQ:
            return HeartbeatReq.unpack(remaining)
        elif message_type is MessageType.HEARTBEAT:
            return Heartbeat.unpack(remaining)
        elif message_type is MessageType.EP_STATUS_REPORT:
            return EPStatusReport.unpack(remaining)
        elif message_type is MessageType.MANAGER_STATUS_REPORT:
            return ManagerStatusReport.unpack(remaining)
        elif message_type is MessageType.TASK:
            return Task.unpack(remaining)
        elif message_type is MessageType.RESULTS_ACK:
            return ResultsAck.unpack(remaining)
        elif message_type is MessageType.TASK_CANCEL:
            return TaskCancel.unpack(remaining)
        elif message_type is MessageType.BAD_COMMAND:
            return BadCommand.unpack(remaining)

        raise Exception(f"Unknown Message Type Code: {message_type}")

    @abstractmethod
    def pack(self):
        raise NotImplementedError()


class Task(Message):
    """
    Task message from the forwarder->interchange
    """

    type = MessageType.TASK

    def __init__(
        self, task_id: str, container_id: str, task_buffer: str | bytes, raw_buffer=None
    ):
        super().__init__()
        self.task_id = task_id
        self.container_id = container_id
        self.task_buffer = task_buffer
        self.raw_buffer = raw_buffer

    def pack(self) -> bytes:
        if self.raw_buffer is None:
            # a type:ignore is needed here
            # task_buffer might be a str  or it might be a bytes (see `unpack`)
            # and we can't tell which is correct
            #
            # rather than thinking hard, preserve the exact current runtime behavior
            #
            # all of this code is going to be eliminated soonish by
            # globus_compute_common.messagepack in part because of issues like this
            add_ons = (
                f"TID={self.task_id};CID={self.container_id};"  # type: ignore
                f"{self.task_buffer}"
            )
            self.raw_buffer = add_ons.encode("utf-8")

        return self.type.pack() + self.raw_buffer

    @classmethod
    def unpack(cls, raw_buffer: bytes):
        b_tid, b_cid, task_buf = raw_buffer.decode("utf-8").split(";", 2)
        return cls(
            b_tid[4:], b_cid[4:], task_buf.encode("utf-8"), raw_buffer=raw_buffer
        )

    def set_local_container(self, container_id):
        self.local_container = container_id


class HeartbeatReq(Message):
    """
    Synchronous request for a Heartbeat.

    This is sent from the Forwarder to the endpoint on start to get an initial
    connection and ensure liveness.
    """

    type = MessageType.HEARTBEAT_REQ

    @property
    def header(self):
        return None

    @property
    def payload(self):
        return None

    @classmethod
    def unpack(cls, msg):
        return cls()

    def pack(self):
        return self.type.pack()


class Heartbeat(Message):
    """
    Generic Heartbeat message, sent in both directions between Forwarder and
    Interchange.
    """

    type = MessageType.HEARTBEAT

    def __init__(self, endpoint_id):
        super().__init__()
        self.endpoint_id = endpoint_id

    @classmethod
    def unpack(cls, msg):
        return cls(msg.decode("ascii"))

    def pack(self):
        return self.type.pack() + self.endpoint_id.encode("ascii")


class EPStatusReport(Message):
    """
    Status report for an endpoint, sent from Interchange to Forwarder.

    Includes EP-wide info such as utilization, as well as per-task status information.
    """

    type = MessageType.EP_STATUS_REPORT

    def __init__(self, endpoint_id, global_state, task_statuses):
        super().__init__()
        self._header = uuid.UUID(endpoint_id).bytes
        self.global_state = global_state
        self.task_statuses = task_statuses

    def __repr__(self):
        name = type(self).__name__
        ep_id = uuid.UUID(bytes=self._header)
        return f"{name}(ep:{ep_id}; task count:{len(self.task_statuses)})"

    @classmethod
    def unpack(cls, msg):
        endpoint_id = str(uuid.UUID(bytes=msg[:16]))
        msg = msg[16:]
        jsonified = msg.decode("ascii")
        global_state, statuses = json.loads(jsonified)
        task_statuses = defaultdict(list)
        for tid, tt in statuses.items():
            for trans in tt:
                task_statuses[tid].append(
                    TaskTransition(
                        timestamp=trans["timestamp"],
                        actor=trans["actor"],
                        state=trans["state"],
                    )
                )
        return cls(endpoint_id, global_state, task_statuses)

    def pack(self):
        statuses = {}
        for tid, tt in self.task_statuses.items():
            for status in tt:
                statuses[tid] = statuses.get(tid, [])
                statuses[tid].append(status.to_dict())
        jsonified = json.dumps([self.global_state, statuses])
        return self.type.pack() + self._header + jsonified.encode("ascii")


class ManagerStatusReport(Message):
    """
    Status report sent from the Manager to the Interchange, which mostly just amounts
    to saying which tasks are now RUNNING.
    """

    type = MessageType.MANAGER_STATUS_REPORT

    def __init__(self, task_statuses, container_switch_count):
        super().__init__()
        self.task_statuses = task_statuses
        self.container_switch_count = container_switch_count

    @classmethod
    def unpack(cls, msg):
        container_switch_count = int.from_bytes(msg[:10], "little")
        msg = msg[10:]
        jsonified = msg.decode("ascii")
        statuses = json.loads(jsonified)
        task_statuses = defaultdict(list)
        for tid, tt in statuses.items():
            for trans in tt:
                task_statuses[tid].append(
                    TaskTransition(
                        timestamp=trans["timestamp"],
                        actor=trans["actor"],
                        state=trans["state"],
                    )
                )
        return cls(task_statuses, container_switch_count)

    def pack(self):
        # TODO: do better than JSON?
        statuses = {}
        for tid, tt in self.task_statuses.items():
            for status in tt:
                statuses[tid] = statuses.get(tid, [])
                statuses[tid].append(status.to_dict())

        jsonified = json.dumps(statuses)
        return (
            self.type.pack()
            + self.container_switch_count.to_bytes(10, "little")
            + jsonified.encode("ascii")
        )


class ResultsAck(Message):
    """
    Results acknowledgement to acknowledge a task result was received by
    the forwarder. Sent from forwarder->interchange
    """

    type = MessageType.RESULTS_ACK

    def __init__(self, task_id):
        super().__init__()
        self.task_id = task_id

    @classmethod
    def unpack(cls, msg):
        return cls(msg.decode("ascii"))

    def pack(self):
        return self.type.pack() + self.task_id.encode("ascii")


class TaskCancel(Message):
    """
    Synchronous request for to cancel a Task.

    This is sent from the Executor to the Interchange
    """

    type = MessageType.TASK_CANCEL

    def __init__(self, task_id):
        super().__init__()
        self.task_id = task_id

    @classmethod
    def unpack(cls, msg):
        return cls(json.loads(msg.decode("ascii")))

    def pack(self):
        return self.type.pack() + json.dumps(self.task_id).encode("ascii")


class BadCommand(Message):
    """
    Error message send to indicate that a command is either
    unknown, malformed or unsupported.
    """

    type = MessageType.BAD_COMMAND

    def __init__(self, reason: str):
        super().__init__()
        self.reason = reason

    @classmethod
    def unpack(cls, msg):
        return cls(msg.decode("ascii"))

    def pack(self):
        return self.type.pack() + self.reason.encode("ascii")



import queue
import random
import uuid

import pytest
from globus_compute_common import messagepack
from globus_compute_endpoint.engines import HighThroughputEngine
from globus_compute_sdk.serialize import ComputeSerializer
from tests.utils import double, ez_pack_function


@pytest.fixture
def engine(tmp_path):
    ep_id = uuid.uuid4()
    engine = HighThroughputEngine(
        label="HTEXEngine",
        heartbeat_period=1,
        worker_debug=True,
    )
    results_queue = queue.Queue()
    engine.start(endpoint_id=ep_id, run_dir=tmp_path, results_passthrough=results_queue)

    yield engine
    engine.shutdown()


def test_engine_submit(engine):
    q = engine.results_passthrough
    task_id = uuid.uuid1()
    serializer = ComputeSerializer()
    task_arg = random.randint(1, 1000)
    task_body = ez_pack_function(serializer, double, (task_arg,), {})
    task_message = messagepack.pack(
        messagepack.message_types.Task(
            task_id=task_id, container_id=uuid.uuid1(), task_buffer=task_body
        )
    )
    future = engine.submit(str(task_id), task_message)
    packed_result = future.result()

    # Confirm that the future got the right answer
    assert isinstance(packed_result, bytes)
    result = messagepack.unpack(packed_result)
    assert isinstance(result, messagepack.message_types.Result)
    assert result.task_id == task_id

    # Confirm that the same result got back though the queue
    for _i in range(10):
        q_msg = q.get(timeout=5)
        assert isinstance(q_msg, dict)

        packed_result_q = q_msg["message"]
        result = messagepack.unpack(packed_result_q)
        # Handle a sneaky EPStatusReport that popped in ahead of the result
        if isinstance(result, messagepack.message_types.EPStatusReport):
            continue

        # At this point the message should be the result
        assert (
            packed_result == packed_result_q
        ), "Result from passthrough_q and future should match"

        assert result.task_id == task_id
        final_result = serializer.deserialize(result.data)
        expected = task_arg * 2
        assert final_result == expected, f"Expected {expected}, but got: {final_result}"
        break



import logging
import queue
import threading
import time
import typing as t
import uuid
from abc import ABC, abstractmethod
from concurrent.futures import Future

from globus_compute_common import messagepack
from globus_compute_common.messagepack.message_types import (
    EPStatusReport,
    Result,
    TaskTransition,
)
from globus_compute_common.tasks import ActorName, TaskState
from globus_compute_endpoint.engines.helper import execute_task
from globus_compute_endpoint.exception_handling import (
    get_error_string,
    get_result_error_details,
)

logger = logging.getLogger(__name__)


class ReportingThread:
    def __init__(
        self, target: t.Callable, args: t.List, reporting_period: float = 30.0
    ):
        """This class wraps threading.Thread to run a callable in a loop
        periodically until the user calls `stop`. A status attribute can
        report exceptions to the parent thread upon failure.
        Parameters
        ----------
        target: Target function to be invoked to get report and post to queue
        args: args to be passed to target fn
        kwargs: kwargs to be passed to target fn
        reporting_period
        """
        self.status: Future = Future()
        self._shutdown_event = threading.Event()
        self.reporting_period = reporting_period
        self._thread = threading.Thread(
            target=self.run_in_loop, args=[target] + args, name="GCReportingThread"
        )

    def start(self):
        logger.info("Start called")
        self._thread.start()

    def run_in_loop(self, target: t.Callable, *args) -> None:
        while True:
            try:
                target(*args)
            except Exception as e:
                # log and update future before exiting, if it is not already set
                self.status.set_exception(exception=e)
                self._shutdown_event.set()
            if self._shutdown_event.wait(timeout=self.reporting_period):
                break

        logger.warning("ReportingThread exiting")

    def stop(self) -> None:
        self._shutdown_event.set()
        self._thread.join(timeout=0.1)


class GlobusComputeEngineBase(ABC):
    """Shared functionality and interfaces required by all GlobusCompute Engines.
    This is designed to plug-in executors following the concurrent.futures.Executor
    interface as execution backends to GlobusCompute
    """

    def __init__(
        self,
        *args: object,
        heartbeat_period_s: float = 30.0,
        endpoint_id: t.Optional[uuid.UUID] = None,
        **kwargs: object,
    ):
        self._shutdown_event = threading.Event()
        self._heartbeat_period_s = heartbeat_period_s
        self.endpoint_id = endpoint_id

        # remove these unused vars that we are adding to just keep
        # endpoint interchange happy
        self.container_type: t.Optional[str] = None
        self.funcx_service_address: t.Optional[str] = None
        self.run_dir: t.Optional[str] = None
        # This attribute could be set by the subclasses in their
        # start method if another component insists on owning the queue.
        self.results_passthrough: queue.Queue[
            dict[str, bytes | str | None]
        ] = queue.Queue()

    @abstractmethod
    def start(
        self,
        *args,
        **kwargs,
    ) -> None:
        raise NotImplementedError

    @abstractmethod
    def get_status_report(self) -> EPStatusReport:
        raise NotImplementedError

    def report_status(self) -> None:
        status_report = self.get_status_report()
        packed: bytes = messagepack.pack(status_report)
        self.results_passthrough.put({"message": packed})

    def _status_report(
        self, shutdown_event: threading.Event, heartbeat_period_s: float
    ):
        while not shutdown_event.wait(timeout=heartbeat_period_s):
            status_report = self.get_status_report()
            packed = messagepack.pack(status_report)
            self.results_passthrough.put({"message": packed})

    def _setup_future_done_callback(self, task_id: str, future: Future) -> None:
        """
        Set up the done() callback for the provided future.

        The done callback handles
        Callback to post result to the passthrough queue
        Parameters
        ----------
        future: Future for which the callback is triggerd
        """

        exec_beg = TaskTransition(  # Reminder: used by *closure*, below
            timestamp=time.time_ns(),
            actor=ActorName.INTERCHANGE,
            state=TaskState.WAITING_FOR_LAUNCH,
        )

        def _done_cb(f: Future):
            if f.exception():
                exc = f.exception()
                code, user_message = get_result_error_details(exc)
                error_details = {"code": code, "user_message": user_message}
                exec_end = TaskTransition(
                    timestamp=time.time_ns(),
                    actor=ActorName.INTERCHANGE,
                    state=TaskState.EXEC_END,
                )
                result_message = dict(
                    task_id=task_id,
                    data=get_error_string(exc=exc),
                    exception=get_error_string(exc=exc),
                    error_details=error_details,
                    task_statuses=[exec_beg, exec_end],  # only transition info we have
                )
                packed = messagepack.pack(Result(**result_message))
            else:
                packed = f.result()

            self.results_passthrough.put({"task_id": task_id, "message": packed})

        future.add_done_callback(_done_cb)

    @abstractmethod
    def _submit(
        self,
        func: t.Callable,
        *args: t.Any,
        **kwargs: t.Any,
    ) -> Future:
        """Subclass should use the internal execution system to implement this"""
        raise NotImplementedError()

    def submit(self, task_id: str, packed_task: bytes) -> Future:
        """GC Endpoints should submit tasks via this method so that tasks are
        tracked properly.
        Parameters
        ----------
        packed_task: messagepack bytes buffer
        Returns
        -------
        future
        """

        future = self._submit(execute_task, task_id, packed_task)
        self._setup_future_done_callback(task_id, future)
        return future

    @abstractmethod
    def shutdown(self, /, **kwargs) -> None:
        raise NotImplementedError()



import os
import pickle
import queue
import shutil

import pytest
from globus_compute_endpoint.engines.high_throughput.manager import Manager
from globus_compute_endpoint.engines.high_throughput.messages import Task


class TestManager:
    @pytest.fixture(autouse=True)
    def test_setup_teardown(self):
        os.makedirs(os.path.join(os.getcwd(), "mock_uid"))
        yield
        shutil.rmtree(os.path.join(os.getcwd(), "mock_uid"))

    def test_remove_worker_init(self, mocker):
        # zmq is being mocked here because it was making tests hang
        mocker.patch(
            "globus_compute_endpoint.engines.high_throughput.manager.zmq.Context"  # noqa: E501
        )

        manager = Manager(logdir="./", uid="mock_uid")
        manager.worker_map.to_die_count["RAW"] = 0
        manager.task_queues["RAW"] = queue.Queue()

        manager.remove_worker_init("RAW")
        task = manager.task_queues["RAW"].get()
        assert isinstance(task, Task)
        assert task.task_id == "KILL"
        assert task.task_buffer == "KILL"

    def test_poll_funcx_task_socket(self, mocker):
        # zmq is being mocked here because it was making tests hang
        mocker.patch(
            "globus_compute_endpoint.engines.high_throughput.manager.zmq.Context"  # noqa: E501
        )

        mock_worker_map = mocker.patch(
            "globus_compute_endpoint.engines.high_throughput.manager.WorkerMap"
        )

        manager = Manager(logdir="./", uid="mock_uid")
        manager.task_queues["RAW"] = queue.Queue()
        manager.logdir = "./"
        manager.worker_type = "RAW"
        manager.worker_procs["0"] = "proc"

        manager.funcx_task_socket.recv_multipart.return_value = (
            b"0",
            b"REGISTER",
            pickle.dumps({"worker_type": "RAW"}),
        )
        manager.poll_funcx_task_socket(test=True)
        mock_worker_map.return_value.register_worker.assert_called_with(b"0", "RAW")

        manager.funcx_task_socket.recv_multipart.return_value = (
            b"0",
            b"WRKR_DIE",
            pickle.dumps(None),
        )
        manager.poll_funcx_task_socket(test=True)
        mock_worker_map.return_value.remove_worker.assert_called_with(b"0")
        assert len(manager.worker_procs) == 0
